""" MatteFormer
Copyright (c) 2022-present NAVER Webtoon
Apache-2.0
"""

import cv2
import random
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint as checkpoint
import numpy as np
from timm.models.layers import DropPath, to_2tuple, trunc_normal_
from ..ops import SpectralNorm


def window_partition(x, window_size, priors=None):
    """
    Args:
        x: (B, H, W, C)
        window_size (int): window size

    Returns:
        windows: (num_windows*B, window_size, window_size, C)
    """

    if priors is not None:
        B, H, W, C = x.shape
        x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
        windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size * window_size, C)

        return_priors = list()
        for p, prior in enumerate(priors):
            prior['unknown'] = prior['unknown'].unsqueeze(1)
            prior['fg'] = prior['fg'].unsqueeze(1)
            prior['bg'] = prior['bg'].unsqueeze(1)
            return_priors.append(prior)

        return windows, return_priors

    else:
        B, H, W, C = x.shape
        x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
        windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)

        return windows


def window_reverse(windows, window_size, H, W):
    """
    Args:
        windows: (num_windows*B, window_size, window_size, C)
        window_size (int): Window size
        H (int): Height of image
        W (int): Width of image

    Returns:
        x: (B, H, W, C)
    """
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    x = windows.view(B, H // window_size, W // window_size, window_size, window_size, -1)
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
    return x


class Mlp(nn.Module):
    """ Multilayer perceptron."""

    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class PatchMerging(nn.Module):
    """ Patch Merging Layer

    Args:
        dim (int): Number of input channels.
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """
    def __init__(self, dim, norm_layer=nn.LayerNorm):
        super().__init__()
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = norm_layer(4 * dim)

    def forward(self, x, H, W):
        """ Forward function.

        Args:
            x: Input feature, tensor size (B, H*W, C).
            H, W: Spatial resolution of the input feature.
        """
        B, L, C = x.shape
        assert L == H * W, "input feature has wrong size"

        x = x.view(B, H, W, C)

        # padding
        pad_input = (H % 2 == 1) or (W % 2 == 1)
        if pad_input:
            x = F.pad(x, (0, 0, 0, W % 2, 0, H % 2))

        x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
        x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
        x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
        x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C
        x = torch.cat([x0, x1, x2, x3], -1)  # B H/2 W/2 4*C
        x = x.view(B, -1, 4 * C)  # B H/2*W/2 4*C

        x = self.norm(x)
        x = self.reduction(x)

        return x


class PatchEmbed(nn.Module):
    """ Image to Patch Embedding

    Args:
        patch_size (int): Patch token size. Default: 4.
        in_chans (int): Number of input image channels. Default: 3.
        embed_dim (int): Number of linear projection output channels. Default: 96.
        norm_layer (nn.Module, optional): Normalization layer. Default: None
    """

    def __init__(self, patch_size=4, in_chans=3, embed_dim=96, norm_layer=None):
        super().__init__()
        patch_size = to_2tuple(patch_size)
        self.patch_size = patch_size

        self.in_chans = in_chans
        self.embed_dim = embed_dim

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)
        if norm_layer is not None:
            self.norm = norm_layer(embed_dim)
        else:
            self.norm = None

    """Forward function."""
    def forward(self, x):

        _, _, H, W = x.size()

        # padding
        if W % self.patch_size[1] != 0:
            x = F.pad(x, (0, self.patch_size[1] - W % self.patch_size[1]))
        if H % self.patch_size[0] != 0:
            x = F.pad(x, (0, 0, 0, self.patch_size[0] - H % self.patch_size[0]))

        x = self.proj(x)  # B C Wh Ww
        if self.norm is not None:
            Wh, Ww = x.size(2), x.size(3)
            x = x.flatten(2).transpose(1, 2)
            x = self.norm(x)
            x = x.transpose(1, 2).view(-1, self.embed_dim, Wh, Ww)

        return x


class PAWSA(nn.Module):
    def __init__(self, dim, window_size, num_heads, qkv_bias=True, qk_scale=None,
                 attn_drop=0., proj_drop=0., use_prior=True, num_priorset=1):

        super().__init__()
        self.dim = dim
        self.window_size = window_size  # Wh, Ww
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        self.use_prior = use_prior
        self.num_priorset = num_priorset

        # use 3 prior vector
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1) + 3*self.num_priorset, num_heads))

        # get pair-wise relative position index for each token inside the window
        coords_h = torch.arange(self.window_size[0])
        coords_w = torch.arange(self.window_size[1])
        coords = torch.stack(torch.meshgrid([coords_h, coords_w]))
        coords_flatten = torch.flatten(coords, 1)
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()
        relative_coords[:, :, 0] += self.window_size[0] - 1
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
        relative_position_index = relative_coords.sum(-1)

        # relative position index
        num_patch = self.window_size[0] * self.window_size[1]
        last_idx = 169
        relative_position_index_withPrior = torch.ones((num_patch, num_patch + self.num_priorset*3)).long() * last_idx
        relative_position_index_withPrior[:num_patch, :num_patch] = relative_position_index
        for i in range(self.num_priorset):
            for j in range(3):
                relative_position_index_withPrior[:, num_patch + i*3 + j:num_patch + i*3 +j +1] = last_idx + i*3 + j
        self.register_buffer("relative_position_index", relative_position_index_withPrior)

        # params
        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)

        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        trunc_normal_(self.relative_position_bias_table, std=.02)

        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, mask=None, _inp_shape=None, prior_bins=None):
        # get shape
        B_, N, C = x.shape
        B, H, W, C = _inp_shape

        # for prior-memory
        prior_highway = None
        for prior in prior_bins:
            uk_prior, fg_prior, bg_prior = prior['unknown'], prior['fg'], prior['bg']
            # prior highways (for shortcut)
            if prior_highway is None:
                prior_highway = torch.cat((uk_prior, fg_prior, bg_prior), dim=1)
            else:
                prior_highway = torch.cat((prior_highway, uk_prior, fg_prior, bg_prior), dim=1)
            # concate prior-tokens to spatial-tokens(x)
            uk_prior = uk_prior.expand((-1, (H // self.window_size[0]) * (W // self.window_size[1]), -1)).unsqueeze(2).contiguous().view(-1, 1, C)
            fg_prior = fg_prior.expand((-1, (H // self.window_size[0]) * (W // self.window_size[1]), -1)).unsqueeze(2).contiguous().view(-1, 1, C)
            bg_prior = bg_prior.expand((-1, (H // self.window_size[0]) * (W // self.window_size[1]), -1)).unsqueeze(2).contiguous().view(-1, 1, C)
            x = torch.cat((x, uk_prior, fg_prior, bg_prior), dim=1)

        relative_position_bias = self.relative_position_bias_table[self.relative_position_index.view(-1)] \
            .view(self.window_size[0] * self.window_size[1], self.window_size[0] * self.window_size[1] + self.num_priorset*3, -1)
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()

        # qkv projection for x
        qkv = self.qkv(x).reshape(B_, N + self.num_priorset*3, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]
        q = q[:, :, :self.window_size[0] * self.window_size[1], :] * self.scale

        # get self-attention features
        attn = (q @ k.transpose(-2, -1))
        attn = attn + relative_position_bias.unsqueeze(0)
        if mask is not None:
            nW = mask.shape[0]
            mask_withPrior = F.pad(mask, (0, self.num_priorset*3, 0, 0))
            attn = attn.view(B_ // nW, nW, self.num_heads, N, N + self.num_priorset*3) + mask_withPrior.contiguous().unsqueeze(1).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, N,  N + self.num_priorset*3)
        attn = self.softmax(attn)

        x = (attn @ v).transpose(1, 2).reshape(B_, N, C)

        # projection
        x = self.proj(x)
        x = self.proj_drop(x)

        # prior projection for prior-memory
        prior_highway = self.qkv(prior_highway).reshape(B, self.num_priorset*3, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        _q_hw, _k_hw, _v_hw = prior_highway[0], prior_highway[1], prior_highway[2]

        _v_hw = _v_hw.transpose(1,2).reshape(B, self.num_priorset*3, C)
        _v_hw = self.proj(_v_hw)
        _v_hw = self.proj_drop(_v_hw)

        return x, _v_hw


class PASTBlock(nn.Module):
    def __init__(self, layer_idx, block_idx, dim, num_heads, window_size=7, shift_size=0,
                 mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0., drop_path=0.,
                 act_layer=nn.GELU, norm_layer=nn.LayerNorm, use_prior=True, num_priorset=0):
        super().__init__()
        self.layer_idx = layer_idx
        self.block_idx = block_idx
        self.dim = dim
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio
        self.use_prior = use_prior
        self.num_priorset = num_priorset

        assert 0 <= self.shift_size < self.window_size, "shift_size must in 0-window_size"

        self.norm1 = norm_layer(dim)
        self.attn = PAWSA(
            dim, window_size=to_2tuple(self.window_size), num_heads=num_heads,
            qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop, use_prior=self.use_prior, num_priorset=self.num_priorset)

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        self.H = None
        self.W = None

    # swin transformer block
    def forward(self, x, mask_matrix, prior_memory=None, area_fg=None, area_bg=None, area_uk=None):
        B, L, C = x.shape
        H, W = self.H, self.W
        assert L == H * W, "input feature has wrong size"

        # shortcut features
        shortcut = x

        # shortcut prior features (current prior + previous blk prior)
        if True:
            shortcut_prior = dict()
            shortcut_ = shortcut.view(B, H, W, C) # [1, 256, 512, 96]

            epsilon = 1e-9
            uk_prior = (area_uk * shortcut_).permute(0,3,1,2)
            uk_prior = uk_prior.sum(axis=(-1,-2)) / (area_uk.sum(axis=(1,2)) + epsilon)
            fg_prior = (area_fg * shortcut_).permute(0, 3, 1, 2)
            fg_prior = fg_prior.sum(axis=(-1,-2)) / (area_fg.sum(axis=(1,2)) + epsilon)
            bg_prior = (area_bg * shortcut_).permute(0, 3, 1, 2)
            bg_prior = bg_prior.sum(axis=(-1, -2)) / (area_bg.sum(axis=(1,2)) + epsilon)

            shortcut_prior['fg'] = fg_prior
            shortcut_prior['bg'] = bg_prior
            shortcut_prior['unknown'] = uk_prior

            # current blk prior on last index
            prior_memory.append(shortcut_prior)

        # forward
        x = self.norm1(x)
        x = x.view(B, H, W, C)

        # forward all prior to norm1
        if True:
            prior_bins = list()
            for p, prior in enumerate(prior_memory):
                prior_bin = dict()
                prior_bin['fg'] = self.norm1(prior['fg'])
                prior_bin['bg'] = self.norm1(prior['bg'])
                prior_bin['unknown'] = self.norm1(prior['unknown'])
                prior_bins.append(prior_bin)

        # pad feature maps to multiples of window size
        pad_l = pad_t = 0
        pad_r = (self.window_size - W % self.window_size) % self.window_size
        pad_b = (self.window_size - H % self.window_size) % self.window_size
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b)) # [1, 259, 518, 96]
        _, Hp, Wp, _ = x.shape

        # cyclic shift
        if self.shift_size > 0:
            shifted_x = torch.roll(x, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2))
            attn_mask = mask_matrix
        else:
            shifted_x = x
            attn_mask = None

        # partition windows
        x_windows, prior_bins = window_partition(shifted_x, self.window_size, prior_bins)
        x_windows = x_windows.view(-1, self.window_size * self.window_size, C)

        # foward attention layer
        attn_windows, priors_after_attn = self.attn(x_windows, mask=attn_mask, prior_bins=prior_bins, _inp_shape=shifted_x.shape)

        # merge windows
        attn_windows = attn_windows.view(-1, self.window_size, self.window_size, C)
        shifted_x = window_reverse(attn_windows, self.window_size, Hp, Wp)  # B H' W' C

        # reverse cyclic shift
        if self.shift_size > 0:
            x = torch.roll(shifted_x, shifts=(self.shift_size, self.shift_size), dims=(1, 2))
        else:
            x = shifted_x

        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()

        x = x.view(B, H * W, C)

        # FFN : second module
        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        if True:
            priors_after_attn = priors_after_attn.chunk(chunks=self.num_priorset*3, axis=1)
            prior_bins_outs = list()
            for idx in range(self.num_priorset):
                prior_bin_out = dict()
                prior_bin_out['unknown'] = priors_after_attn[idx*3+0]
                prior_bin_out['fg'] = priors_after_attn[idx*3+1]
                prior_bin_out['bg'] = priors_after_attn[idx*3+2]
                prior_bins_outs.append(prior_bin_out)

            # FFN : prior second module
            prior_memory_out = list()
            for p, (prior_shortcut, prior_out) in enumerate(zip(prior_memory, prior_bins_outs)):
                prior_out_uk = prior_shortcut['unknown'] + self.drop_path(prior_out['unknown'].squeeze())
                prior_out_uk = prior_out_uk + self.drop_path(self.mlp(self.norm2(prior_out_uk)))

                prior_out_fg = prior_shortcut['fg'] + self.drop_path(prior_out['fg'].squeeze())
                prior_out_fg = prior_out_fg + self.drop_path(self.mlp(self.norm2(prior_out_fg)))

                prior_out_bg = prior_shortcut['bg'] + self.drop_path(prior_out['bg'].squeeze())
                prior_out_bg = prior_out_bg + self.drop_path(self.mlp(self.norm2(prior_out_bg)))

                prior_out = {'fg':prior_out_fg, 'bg':prior_out_bg, 'unknown':prior_out_uk}
                prior_memory_out.append(prior_out)

        return x, prior_memory_out


class BasicLayer(nn.Module):

    def __init__(self, layer_idx,
                 dim, depth, num_heads, window_size=7, mlp_ratio=4., qkv_bias=True,
                 qk_scale=None, drop=0., attn_drop=0., drop_path=0., norm_layer=nn.LayerNorm, downsample=None,
                 use_checkpoint=False, use_prior=True):
        super().__init__()
        self.layer_idx=layer_idx
        self.window_size = window_size
        self.shift_size = window_size // 2
        self.depth = depth
        self.use_checkpoint = use_checkpoint

        # build blocks
        self.blocks = nn.ModuleList([
            PASTBlock(
                layer_idx=layer_idx,
                block_idx=i,
                dim=dim,
                num_heads=num_heads,
                window_size=window_size,
                shift_size=0 if (i % 2 == 0) else window_size // 2,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop,
                attn_drop=attn_drop,
                drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path,
                norm_layer=norm_layer,
                use_prior=use_prior,
                num_priorset=i+1)
            for i in range(depth)])

        # patch merging layer
        if downsample is not None:
            self.downsample = downsample(dim=dim, norm_layer=norm_layer)
        else:
            self.downsample = None

    def _make_shortcut(self, inplane, planes, norm_layer=nn.BatchNorm2d):
        return nn.Sequential(
            nn.Linear(inplane, planes, bias=False),
            nn.ReLU(inplace=True),
        )

    # basic layer
    def forward(self, x, H, W, area_fg, area_bg, area_uk):

        # calculate attention mask for SW-MSA
        Hp = int(np.ceil(H / self.window_size)) * self.window_size
        Wp = int(np.ceil(W / self.window_size)) * self.window_size
        img_mask = torch.zeros((1, Hp, Wp, 1), device=x.device)  # 1 Hp Wp 1
        h_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        w_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))

        # img_mask -> index map (0~8) # [1, 259, 518, 1]
        cnt = 0
        for h in h_slices:
            for w in w_slices:
                img_mask[:, h, w, :] = cnt
                cnt += 1

        # get attn_mask
        mask_windows = window_partition(img_mask, self.window_size)
        mask_windows = mask_windows.view(-1, self.window_size * self.window_size)

        attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
        attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))

        # forward blocks
        prior_memory = []
        for b, blk in enumerate(self.blocks):
            blk.H, blk.W = H, W

            if self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x, attn_mask)
            else:
                x, prior_memory = blk(x, attn_mask, prior_memory=prior_memory,
                                    area_fg=area_fg, area_bg=area_bg, area_uk=area_uk)

        if self.downsample is not None:
            x_down = self.downsample(x, H, W)
            Wh, Ww = (H + 1) // 2, (W + 1) // 2
            return x, H, W, x_down, Wh, Ww
        else:
            return x, H, W, x, H, W


class MatteFormer(nn.Module):

    def __init__(self,
                 patch_size=4, in_chans=3, embed_dim=96, depths=[2, 2, 6, 2],
                 num_heads=[3, 6, 12, 24], window_size=7, mlp_ratio=4., qkv_bias=True, qk_scale=None,
                 drop_rate=0., attn_drop_rate=0., drop_path_rate=0.2, norm_layer=nn.LayerNorm,
                 patch_norm=True, out_indices=(0, 1, 2, 3), use_checkpoint=False):
        super().__init__()

        self.num_layers = len(depths)
        self.embed_dim = embed_dim
        self.patch_norm = patch_norm
        self.out_indices = out_indices

        # split image into non-overlapping patches (with trimap)
        self.patch_embed = PatchEmbed(
            patch_size=patch_size, in_chans=in_chans+3, embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None)

        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule

        # build layers
        self.layers = nn.ModuleList()
        for i_layer in range(self.num_layers):
            layer = BasicLayer(
                layer_idx = i_layer,
                dim=int(embed_dim * 2 ** i_layer),
                depth=depths[i_layer],
                num_heads=num_heads[i_layer],
                window_size=window_size,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[sum(depths[:i_layer]):sum(depths[:i_layer + 1])],
                norm_layer=norm_layer,
                downsample=PatchMerging if (i_layer < self.num_layers - 1) else None,
                use_checkpoint=use_checkpoint,
                use_prior=True,
            )
            self.layers.append(layer)

        num_features = [int(embed_dim * 2 ** i) for i in range(self.num_layers)]
        self.num_features = num_features

        # add a norm layer for each output [from Swin-Transformer-Semantic-Segmentation]
        for i_layer in out_indices:
            layer = norm_layer(num_features[i_layer])
            layer_name = f'norm{i_layer}'
            self.add_module(layer_name, layer)

        # add shortcut layer [from MG-Matting]
        self.shortcut = nn.ModuleList()
        shortcut_inplanes = [[6, 32], [96, 32], [96, 64], [192, 128], [384, 256], [768, 512]]
        for shortcut in shortcut_inplanes:
            inplane, planes = shortcut
            self.shortcut.append(self._make_shortcut(inplane=inplane, planes=planes, norm_layer=nn.BatchNorm2d))

    def _make_shortcut(self, inplane, planes, norm_layer=nn.BatchNorm2d):
        '''
        came from MGMatting
        '''
        return nn.Sequential(
            SpectralNorm(nn.Conv2d(inplane, planes, kernel_size=3, padding=1, bias=False)),
            nn.ReLU(inplace=True),
            norm_layer(planes),
            SpectralNorm(nn.Conv2d(planes, planes, kernel_size=3, padding=1, bias=False)),
            nn.ReLU(inplace=True),
            norm_layer(planes)
        )

    def forward(self, x, trimapmask, sampleidx=None):
        # set outputs
        outs = []

        # get outs[0]
        outs.append(self.shortcut[0](x))

        # forward patch-embedding layer
        x = self.patch_embed(x)
        _, _, Wh, Ww = x.shape
        trimapmask = F.interpolate(trimapmask, scale_factor=1/4, mode='nearest')

        # get outs[1]
        outs.append(self.shortcut[1](F.upsample_bilinear(x, scale_factor=2.0)))

        # dropout
        x = x.flatten(2).transpose(1, 2)
        x = self.pos_drop(x)

        # get outs[2~5]
        for i in range(self.num_layers):
            layer = self.layers[i]

            trimapmask_ = F.interpolate(trimapmask, scale_factor=1/(pow(2, i)), mode='nearest')
            area_fg = trimapmask_[:, 0, :, :].unsqueeze(-1)  # background area
            area_bg = trimapmask_[:, 2, :, :].unsqueeze(-1)  # foreground area
            area_uk = trimapmask_[:, 1, :, :].unsqueeze(-1)  # unknown area

            x_out, H, W, x, Wh, Ww = layer(x, Wh, Ww, area_fg=area_fg, area_bg=area_bg, area_uk=area_uk)

            if i in self.out_indices:
                norm_layer = getattr(self, f'norm{i}')
                x_out = norm_layer(x_out)
                out = x_out.view(-1, H, W, self.num_features[i]).permute(0, 3, 1, 2).contiguous()
                out = self.shortcut[i+2](out)
                outs.append(out)

        return tuple(outs)

    def train(self, mode=True):
        super(MatteFormer, self).train(mode)


if __name__ == '__main__':
    model = MatteFormer().cuda()
    print(model)

    out = model(torch.ones(2, 6, 512, 512).cuda(), torch.ones(2,3,512,512).cuda())
    print(len(out))

    print('MODEL DEBUG')


import os
import sys

import numpy as np
import torch
import torch.nn as nn

from .nets import MyGenerator
from .util import set_random_colors, plot_kpts, plot_kpt
from .vis_utils.vis_68 import contour_lines_68
from .vis_utils.vis_756 import contour_lines_756
from ..utils import check_ckpt_exist, get_url_id

class Face_Reshaper(nn.Module):
    def __init__(self, img_size : int = 1024, color_num : int = 68, folder_name='face_reshaper', ckpt_name = 'G_1024_28500.pt', force=False, device = 'cuda'):
        """
            Related Links
            --------
            https://meta-portrait.github.io/
            
            Options
            --------
            ckpt_name
                - G_13000.pt : image size(512)
            

            
            get_vis_color_landmark
            --------
            - input
                - lmks
                    - shape : (68 2) or (756,2)
                    
                - size
                    - dtype : int
                    - default : 512
                    - plz check landmark size

                - color_nums
                    - dtype : int
                    - default : 68 or 756
            
            - output
                - dicts 'vis_color_lmk'
                    - dtype : numpy
                    - shape : (3, h, w)
                    - min max : (0 255)
            
            Forwards
            --------
            - input
                - from face
                    - aligned face with ffhq rules
                    - dtype : tensor
                    - shape : (b, 3, h, w)
                    - min max : (-1 1)
                
                - from landmark vis
                    - dtype : tensor
                    - shape : (b, 3, h , w)
                    - min max : (-1 1)
                    
                - to landmark vis
                    - dtype : tensor
                    - shape : (b, 3, h , w)
                    - min max : (-1 1)
            - output
                - dicts 'result'
                    - dtype : tensor
                    - shape : (b, 3, h, w)
                    - min max : (-1 1)
                
        """
        super(Face_Reshaper, self).__init__()
        self.device = device
        self.img_size = img_size
        self.generator = MyGenerator().to(self.device)
        
        url_id = get_url_id('~/.invz_pack/', folder_name, ckpt_name)
        root = os.path.join('~/.invz_pack/', folder_name)
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        ckpt = torch.load(ckpt_path, map_location=self.device)
        
        self.generator.load_state_dict(ckpt['model'])
        for param in self.generator.parameters():
            param.requires_grad = False
        self.generator.eval()
        del ckpt
        
        if color_num == 68:
            self.contour_lines = np.array(contour_lines_68)
            self.colors = set_random_colors(num=color_num)
        else:
            self.contour_lines = np.array(contour_lines_756)
            self.colors = set_random_colors(num=color_num)
        
    def get_vis_color_landmark(self, lmks, size=512, dicts={}) -> dict:
        canvas = np.zeros((size, size, 3))
        for (st, ed), color in zip(self.contour_lines, self.colors):
            st_point, ed_point = lmks[int(st)], lmks[int(ed)]
            canvas = plot_kpt(canvas, st_point, ed_point, color)
        dicts['vis_color_lmk'] = canvas
        return dicts

    def forward(self, from_face, from_color_vis_lmk, to_color_vis_lmk, dicts={}) -> dict:
        inputs = torch.cat((from_color_vis_lmk, to_color_vis_lmk), dim=1)
        result, dense_motion = self.generator(from_face, inputs)
        dicts['from_vis_lmks'] = from_color_vis_lmk
        dicts['to_vis_lmks'] = to_color_vis_lmk
        dicts['result'] = result
        dicts['dense_motion'] = dense_motion
        return dicts
        
        
    def _get_vis_color_landmark(self, canvas, lmks, dicts={}) -> dict:
        for (st, ed), color in zip(self.contour_lines, self.colors):
            st_point, ed_point = lmks[int(st)], lmks[int(ed)]
            canvas = plot_kpt(canvas, st_point, ed_point, color, 2)
        dicts['vis_color_lmk'] = canvas
        return dicts


import cv2
from PIL import Image
import numpy as np


    # -*- coding: utf-8 -*-
"""
Created on Tue Jul 11 06:54:28 2017

@author: zhaoyafei
"""

import numpy as np
from numpy.linalg import inv, norm, lstsq
from numpy.linalg import matrix_rank as rank

class MatlabCp2tormException(Exception):
    def __str__(self):
        return 'In File {}:{}'.format(
                __file__, super.__str__(self))

def tformfwd(trans, uv):
    """
    Function:
    ----------
        apply affine transform 'trans' to uv

    Parameters:
    ----------
        @trans: 3x3 np.array
            transform matrix
        @uv: Kx2 np.array
            each row is a pair of coordinates (x, y)

    Returns:
    ----------
        @xy: Kx2 np.array
            each row is a pair of transformed coordinates (x, y)
    """
    uv = np.hstack((
        uv, np.ones((uv.shape[0], 1))
    ))
    xy = np.dot(uv, trans)
    xy = xy[:, 0:-1]
    return xy


def tforminv(trans, uv):
    """
    Function:
    ----------
        apply the inverse of affine transform 'trans' to uv

    Parameters:
    ----------
        @trans: 3x3 np.array
            transform matrix
        @uv: Kx2 np.array
            each row is a pair of coordinates (x, y)

    Returns:
    ----------
        @xy: Kx2 np.array
            each row is a pair of inverse-transformed coordinates (x, y)
    """
    Tinv = inv(trans)
    xy = tformfwd(Tinv, uv)
    return xy


def findNonreflectiveSimilarity(uv, xy, options=None):

    options = {'K': 2}

    K = options['K']
    M = xy.shape[0]
    x = xy[:, 0].reshape((-1, 1))  # use reshape to keep a column vector
    y = xy[:, 1].reshape((-1, 1))  # use reshape to keep a column vector
    # print('--->x, y:\n', x, y

    tmp1 = np.hstack((x, y, np.ones((M, 1)), np.zeros((M, 1))))
    tmp2 = np.hstack((y, -x, np.zeros((M, 1)), np.ones((M, 1))))
    X = np.vstack((tmp1, tmp2))
    # print('--->X.shape: ', X.shape
    # print('X:\n', X

    u = uv[:, 0].reshape((-1, 1))  # use reshape to keep a column vector
    v = uv[:, 1].reshape((-1, 1))  # use reshape to keep a column vector
    U = np.vstack((u, v))
    # print('--->U.shape: ', U.shape
    # print('U:\n', U

    # We know that X * r = U
    if rank(X) >= 2 * K:
        r, _, _, _ = lstsq(X, U)
        r = np.squeeze(r)
    else:
        raise Exception('cp2tform:twoUniquePointsReq')

    # print('--->r:\n', r

    sc = r[0]
    ss = r[1]
    tx = r[2]
    ty = r[3]

    Tinv = np.array([
        [sc, -ss, 0],
        [ss,  sc, 0],
        [tx,  ty, 1]
    ])

    # print('--->Tinv:\n', Tinv

    T = inv(Tinv)
    # print('--->T:\n', T

    T[:, 2] = np.array([0, 0, 1])

    return T, Tinv


def findSimilarity(uv, xy, options=None):

    options = {'K': 2}

#    uv = np.array(uv)
#    xy = np.array(xy)

    # Solve for trans1
    trans1, trans1_inv = findNonreflectiveSimilarity(uv, xy, options)

    # Solve for trans2

    # manually reflect the xy data across the Y-axis
    xyR = xy
    xyR[:, 0] = -1 * xyR[:, 0]

    trans2r, trans2r_inv = findNonreflectiveSimilarity(uv, xyR, options)

    # manually reflect the tform to undo the reflection done on xyR
    TreflectY = np.array([
        [-1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]
    ])

    trans2 = np.dot(trans2r, TreflectY)

    # Figure out if trans1 or trans2 is better
    xy1 = tformfwd(trans1, uv)
    norm1 = norm(xy1 - xy)

    xy2 = tformfwd(trans2, uv)
    norm2 = norm(xy2 - xy)

    if norm1 <= norm2:
        return trans1, trans1_inv
    else:
        trans2_inv = inv(trans2)
        return trans2, trans2_inv


def get_similarity_transform(src_pts, dst_pts, reflective=True):
    """
    Function:
    ----------
        Find Similarity Transform Matrix 'trans':
            u = src_pts[:, 0]
            v = src_pts[:, 1]
            x = dst_pts[:, 0]
            y = dst_pts[:, 1]
            [x, y, 1] = [u, v, 1] * trans

    Parameters:
    ----------
        @src_pts: Kx2 np.array
            source points, each row is a pair of coordinates (x, y)
        @dst_pts: Kx2 np.array
            destination points, each row is a pair of transformed
            coordinates (x, y)
        @reflective: True or False
            if True:
                use reflective similarity transform
            else:
                use non-reflective similarity transform

    Returns:
    ----------
       @trans: 3x3 np.array
            transform matrix from uv to xy
        trans_inv: 3x3 np.array
            inverse of trans, transform matrix from xy to uv
    """

    if reflective:
        trans, trans_inv = findSimilarity(src_pts, dst_pts)
    else:
        trans, trans_inv = findNonreflectiveSimilarity(src_pts, dst_pts)

    return trans, trans_inv


def cvt_tform_mat_for_cv2(trans):
    """
    Function:
    ----------
        Convert Transform Matrix 'trans' into 'cv2_trans' which could be
        directly used by cv2.warpAffine():
            u = src_pts[:, 0]
            v = src_pts[:, 1]
            x = dst_pts[:, 0]
            y = dst_pts[:, 1]
            [x, y].T = cv_trans * [u, v, 1].T

    Parameters:
    ----------
        @trans: 3x3 np.array
            transform matrix from uv to xy

    Returns:
    ----------
        @cv2_trans: 2x3 np.array
            transform matrix from src_pts to dst_pts, could be directly used
            for cv2.warpAffine()
    """
    cv2_trans = trans[:, 0:2].T

    return cv2_trans


def get_similarity_transform_for_cv2(src_pts, dst_pts, reflective=True):
    """
    Function:
    ----------
        Find Similarity Transform Matrix 'cv2_trans' which could be
        directly used by cv2.warpAffine():
            u = src_pts[:, 0]
            v = src_pts[:, 1]
            x = dst_pts[:, 0]
            y = dst_pts[:, 1]
            [x, y].T = cv_trans * [u, v, 1].T

    Parameters:
    ----------
        @src_pts: Kx2 np.array
            source points, each row is a pair of coordinates (x, y)
        @dst_pts: Kx2 np.array
            destination points, each row is a pair of transformed
            coordinates (x, y)
        reflective: True or False
            if True:
                use reflective similarity transform
            else:
                use non-reflective similarity transform

    Returns:
    ----------
        @cv2_trans: 2x3 np.array
            transform matrix from src_pts to dst_pts, could be directly used
            for cv2.warpAffine()
    """
    trans, trans_inv = get_similarity_transform(src_pts, dst_pts, reflective)
    cv2_trans = cvt_tform_mat_for_cv2(trans)
    cv2_trans_inv = cvt_tform_mat_for_cv2(trans_inv)

    return cv2_trans, cv2_trans_inv



import torch
import torch.nn as nn
from torch.nn import Linear, Conv2d, BatchNorm1d, BatchNorm2d, PReLU, ReLU, Sigmoid, Dropout, MaxPool2d, \
    AdaptiveAvgPool2d, Sequential, Module
from collections import namedtuple


class Flatten(Module):
    def forward(self, input):
        return input.view(input.size(0), -1)

def l2_norm(input, axis=1):
    norm = torch.norm(input, 2, axis, True)
    output = torch.div(input, norm)

    return output


class SEModule(Module):
    def __init__(self, channels, reduction):
        super(SEModule, self).__init__()
        self.avg_pool = AdaptiveAvgPool2d(1)
        self.fc1 = Conv2d(
            channels, channels // reduction, kernel_size=1, padding=0, bias=False)

        nn.init.xavier_uniform_(self.fc1.weight.data)

        self.relu = ReLU(inplace=True)
        self.fc2 = Conv2d(
            channels // reduction, channels, kernel_size=1, padding=0, bias=False)

        self.sigmoid = Sigmoid()

    def forward(self, x):
        module_input = x
        x = self.avg_pool(x)
        x = self.fc1(x)
        x = self.relu(x)
        x = self.fc2(x)
        x = self.sigmoid(x)

        return module_input * x


class bottleneck_IR(Module):
    def __init__(self, in_channel, depth, stride):
        super(bottleneck_IR, self).__init__()
        if in_channel == depth:
            self.shortcut_layer = MaxPool2d(1, stride)
        else:
            self.shortcut_layer = Sequential(
                Conv2d(in_channel, depth, (1, 1), stride, bias=False), BatchNorm2d(depth))
        self.res_layer = Sequential(
            BatchNorm2d(in_channel),
            Conv2d(in_channel, depth, (3, 3), (1, 1), 1, bias=False),
            PReLU(depth),
            Conv2d(depth, depth, (3, 3), stride, 1, bias=False),
            BatchNorm2d(depth))

    def forward(self, x):
        shortcut = self.shortcut_layer(x)
        res = self.res_layer(x)

        return res + shortcut


class bottleneck_IR_SE(Module):
    def __init__(self, in_channel, depth, stride):
        super(bottleneck_IR_SE, self).__init__()
        if in_channel == depth:
            self.shortcut_layer = MaxPool2d(1, stride)
        else:
            self.shortcut_layer = Sequential(
                Conv2d(in_channel, depth, (1, 1), stride, bias=False),
                BatchNorm2d(depth))
        self.res_layer = Sequential(
            BatchNorm2d(in_channel),
            Conv2d(in_channel, depth, (3, 3), (1, 1), 1, bias=False),
            PReLU(depth),
            Conv2d(depth, depth, (3, 3), stride, 1, bias=False),
            BatchNorm2d(depth),
            SEModule(depth, 16)
        )

    def forward(self, x):
        shortcut = self.shortcut_layer(x)
        res = self.res_layer(x)

        return res + shortcut


class Bottleneck(namedtuple('Block', ['in_channel', 'depth', 'stride'])):
    '''A named tuple describing a ResNet block.'''


def get_block(in_channel, depth, num_units, stride=2):

    return [Bottleneck(in_channel, depth, stride)] + [Bottleneck(depth, depth, 1) for i in range(num_units - 1)]


def get_blocks(num_layers):
    if num_layers == 50:
        blocks = [
            get_block(in_channel=64, depth=64, num_units=3),
            get_block(in_channel=64, depth=128, num_units=4),
            get_block(in_channel=128, depth=256, num_units=14),
            get_block(in_channel=256, depth=512, num_units=3)
        ]
    elif num_layers == 100:
        blocks = [
            get_block(in_channel=64, depth=64, num_units=3),
            get_block(in_channel=64, depth=128, num_units=13),
            get_block(in_channel=128, depth=256, num_units=30),
            get_block(in_channel=256, depth=512, num_units=3)
        ]
    elif num_layers == 152:
        blocks = [
            get_block(in_channel=64, depth=64, num_units=3),
            get_block(in_channel=64, depth=128, num_units=8),
            get_block(in_channel=128, depth=256, num_units=36),
            get_block(in_channel=256, depth=512, num_units=3)
        ]

    return blocks


class Backbone(Module):
    def __init__(self, num_layers=100):
        super(Backbone, self).__init__()
        blocks = get_blocks(num_layers)
        unit_module = bottleneck_IR
        self.input_layer = Sequential(Conv2d(3, 64, (3, 3), 1, 1, bias=False),
                                      BatchNorm2d(64),
                                      PReLU(64))
        self.output_layer = Sequential(BatchNorm2d(512),
                                       Dropout(0.4),
                                       Flatten(),
                                       Linear(512 * 7 * 7, 512),
                                       BatchNorm1d(512, affine=False))

        modules = []
        for block in blocks:
            for bottleneck in block:
                modules.append(
                    unit_module(bottleneck.in_channel,
                                bottleneck.depth,
                                bottleneck.stride))
        self.body = Sequential(*modules)

    def forward(self, x):
        x = self.input_layer(x)
        x = self.body(x)
        # conv_out = x.view(x.shape[0], -1)
        x = self.output_layer(x)

        return l2_norm(x)


import os
import sys
sys.path.append('../')
cwd = os.path.dirname(os.path.realpath(__file__))

import cv2
import numpy as np
from PIL import Image

import torch
import torch.nn as nn
import torch.nn.functional as F

import torchvision.transforms as transforms

from .nets import Generator_MatteFormer, RAFT
from .nets.utils import InputPadder, warp, get_unknown_tensor_from_pred
from ..utils import check_ckpt_exist, get_url_id

# from utils.util import crop

class Face_Matting(nn.Module):
    def __init__(self, img_size : int = 512, window_size = 7, split_amount=4, folder_name='face_matting', \
                    matt_PATH = 'G_79k.pt', raft_path = 'raft-things.pth', force=False, device = 'cuda'):
        super(Face_Matting, self).__init__()
        
        self.device = device
        
        url_id = get_url_id('~/.invz_pack/', folder_name, matt_PATH)
        root = os.path.join('~/.invz_pack/', folder_name)
        self.matt_PATH = check_ckpt_exist(root, ckpt_name = matt_PATH, url_id = url_id, force = force)
        
        url_id = get_url_id('~/.invz_pack/', folder_name, raft_path)
        root = os.path.join('~/.invz_pack/', folder_name)
        self.raft_path = check_ckpt_exist(root, ckpt_name = raft_path, url_id = url_id, force = force)
        
        self.img_size = img_size
        self.window_size = window_size
        self.side_window = self.window_size//2
        self.split_amount = split_amount
        self.k = cv2.getStructuringElement(cv2.MORPH_RECT, (3,3))
        
        # self.wt = torch.exp(-(torch.arange(self.window_size).float()-self.side_window)**2/(2*((self.side_window+0.5)))).reshape(self.window_size,1,1,1)
        self.wt = torch.exp(-(torch.arange(self.window_size).float()-self.side_window)**2/(2*((self.side_window+0.5)**2))).reshape(self.window_size,1,1,1)
        
        self._model_setting(self.matt_PATH, self.raft_path)
        
        self.tf_color = transforms.Compose([
            transforms.Resize((self.img_size, self.img_size)),
            transforms.ToTensor(),
            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
        ])
        self.kernel = np.ones((3, 3), np.uint8)
        
    def forward(self, image, trimap, dicts={}):
        preds = self.generator(image, trimap)
            
        alpha_pred_os1, alpha_pred_os4, alpha_pred_os8 = preds['alpha_os1'], preds['alpha_os4'], preds['alpha_os8']

        # refinement
        alpha_pred = alpha_pred_os8.clone().detach()
        weight_os4 = get_unknown_tensor_from_pred(alpha_pred, rand_width=30, train_mode=False)
        alpha_pred[weight_os4>0] = alpha_pred_os4[weight_os4>0]
        weight_os1 = get_unknown_tensor_from_pred(alpha_pred, rand_width=15, train_mode=False)
        alpha_pred[weight_os1>0] = alpha_pred_os1[weight_os1>0]

        alpha_pred = alpha_pred[0, 0, ...].data.cpu() * 255
        
        return alpha_pred
    
    def get_trimap(self, label, pad_size=32, dilate_iter=15, erode_iter=10):
        # skin area
        skin_mask = np.where(label < 14, 1, 0) - np.where(label == 7, 1, 0) - np.where(label == 8, 1, 0) - np.where(label == 9, 1, 0) - np.where(label == 0, 1, 0)
        pad_skin_mask = np.pad(skin_mask, ((pad_size,pad_size),(pad_size,pad_size)))

        pred_area = cv2.dilate(np.array(pad_skin_mask).astype(np.float64), self.k, iterations=dilate_iter)
        certain_area = cv2.erode(np.array(pad_skin_mask).astype(np.float64), self.k, iterations=erode_iter)
        
        padded_trimap = pred_area * (1-certain_area) + 2 * certain_area
        trimap = padded_trimap[pad_size:-pad_size,pad_size:-pad_size]
        # trimap = np.concatenate((trimap[:256,:], skin_mask[256:,:]*2), axis=0)
        _trimap = torch.from_numpy(trimap.astype(np.long))
        trimap_onehot = F.one_hot(_trimap, num_classes=3).permute(2, 0, 1).float().unsqueeze(0)
        
        return trimap_onehot
    
    def warping(self, image_chunks, label_chunks, flow_chunks):
        warp_label_chunks = []
        for image_chunk, label_chunk, flow_chunk in zip(image_chunks, label_chunks, flow_chunks):
            warp_label_chunk = []
            pad_image_chunk = self._get_pad_chunk(image_chunk)
            pad_label_chunk = self._get_pad_chunk(label_chunk)
            pad_flow_chunk = self._get_pad_chunk(flow_chunk)
            for index in range(self.side_window, len(pad_image_chunk)-self.side_window):
                images = torch.cat(pad_image_chunk[index-self.side_window:index+self.side_window+1], dim=0)
                labels = torch.cat(pad_label_chunk[index-self.side_window:index+self.side_window+1], dim=0)
                flows = torch.cat(pad_flow_chunk[index-self.side_window:index+self.side_window+1], dim=0)

                output, mask = warp(torch.cat((images, labels), dim=1), flows)

                aligned_Is = output[:,0:3].detach()
                aligned_Ps = output[:,3:].detach()
                _image = pad_image_chunk[index].repeat(2*self.side_window+1, 1, 1, 1)
            
                # the spatial weight
                ws = torch.exp(-((aligned_Is-_image)**2).mean(dim=1, keepdims=True)/(2*(0.2**2))) * mask[:,0:1]
                aligned_Ps[self.side_window] = labels[self.side_window].to(labels.device)
                # the weight between i and i shoud be 1.0
                ws[self.side_window,:,:,:] = 1.0
                weights = ws * self.wt # weights = w(j,i)
                weights = weights / weights.sum(dim=(0), keepdims=True)
                fused_Ps = (aligned_Ps * weights).sum(dim=0, keepdims=True)
                if (aligned_Ps * weights).sum(dim=0, keepdims=True).argmax(1).max() != 0:
                    fused_Ps = fused_Ps.argmax(1).squeeze()
                    fused_Ps = F.one_hot(fused_Ps, num_classes=fused_Ps.max()+1).permute(2, 0, 1).float().unsqueeze(0)
                warp_label_chunk.append(fused_Ps)
            warp_label_chunks.append(warp_label_chunk)
        return warp_label_chunks
            
    def get_flow(self, image_chunks):
        flow_chunks = []
        
        for image_chunk in image_chunks:
            flow_chunk = []
            pad_image_chunk = self._get_pad_chunk(image_chunk)
            for index in range(self.side_window, len(pad_image_chunk)-self.side_window):
                image2 = torch.cat(pad_image_chunk[index-self.side_window:index+self.side_window+1], dim=0).to(self.device)
                image1 = pad_image_chunk[index].repeat(2*self.side_window+1,1,1,1).to(self.device)
                padder = InputPadder(image1.shape)
                
                image1, image2 = padder.pad(image1, image2)
                with torch.no_grad():
                    _, flows_up = self.raft((image1+1)*255.0/2, (image2+1)*255.0/2, iters=20, test_mode=True)
                flow_up = flows_up[self.side_window:self.side_window+1].cpu()
                flow_chunk.append(flow_up)
            flow_chunks.append(flow_chunk)
        return flow_chunks
    
    def _get_chunk_list(self, lists, n):
        m = len(lists)//n + 1 if len(lists)%n != 0 else len(lists)//n
        return [lists[i:i+m] for i in range(0,len(lists),m)]
    
    def _get_pad_chunk(self, chunk):
        pad_chunk = chunk[1:self.side_window+1][::-1] + chunk + chunk[-self.side_window-1:-1][::-1]
        return pad_chunk
    
    def _model_setting(self, face_matting_ckpt_path, raft_ckpt_path):
        # face matting model setting
        self.generator = Generator_MatteFormer().to(self.device)
        ckpt = torch.load(os.path.join(cwd, face_matting_ckpt_path), map_location=self.device)
        self.generator.load_state_dict(ckpt['model'])
        for param in self.generator.parameters():
            param.requires_grad = False
        self.generator.eval()
        
        self.raft = torch.nn.DataParallel(RAFT())
        ckpt = torch.load(os.path.join(cwd, raft_ckpt_path), map_location=self.device)
        self.raft.load_state_dict(ckpt)
        self.raft = self.raft.module
        for param in self.raft.parameters():
            param.requires_grad = False
        self.raft.eval().to(self.device)
        del ckpt
        


# -*- coding: utf-8 -*-
#
# Max-Planck-Gesellschaft zur Förderung der Wissenschaften e.V. (MPG) is
# holder of all proprietary rights on this computer program.
# Using this computer program means that you agree to the terms 
# in the LICENSE file included with this software distribution. 
# Any use not explicitly granted by the LICENSE is prohibited.
#
# Copyright©2019 Max-Planck-Gesellschaft zur Förderung
# der Wissenschaften e.V. (MPG). acting on behalf of its Max Planck Institute
# for Intelligent Systems. All rights reserved.
#
# For comments or questions, please email us at deca@tue.mpg.de
# For commercial licensing contact, please contact ps-license@tuebingen.mpg.de

import numpy as np
import torch.nn as nn
import torch
import torch.nn.functional as F
from . import resnet

class ResnetEncoder(nn.Module):
    def __init__(self, outsize, last_op=None):
        super(ResnetEncoder, self).__init__()
        feature_size = 2048
        self.encoder = resnet.load_ResNet50Model() #out: 2048
        ### regressor
        self.layers = nn.Sequential(
            nn.Linear(feature_size, 1024),
            nn.ReLU(),
            nn.Linear(1024, outsize)
        )
        self.last_op = last_op

    def forward(self, inputs):
        features = self.encoder(inputs)
        parameters = self.layers(features)
        if self.last_op:
            parameters = self.last_op(parameters)
        return parameters



# -*- coding: utf-8 -*-
#
# Max-Planck-Gesellschaft zur Förderung der Wissenschaften e.V. (MPG) is
# holder of all proprietary rights on this computer program.
# Using this computer program means that you agree to the terms 
# in the LICENSE file included with this software distribution. 
# Any use not explicitly granted by the LICENSE is prohibited.
#
# Copyright©2019 Max-Planck-Gesellschaft zur Förderung
# der Wissenschaften e.V. (MPG). acting on behalf of its Max Planck Institute
# for Intelligent Systems. All rights reserved.
#
# For comments or questions, please email us at deca@tue.mpg.de
# For commercial licensing contact, please contact ps-license@tuebingen.mpg.de

import numpy as np
import torch
import torch.nn.functional as F
import math
from collections import OrderedDict
import os
from scipy.ndimage import morphology
from skimage.io import imsave
import cv2
import torchvision

def upsample_mesh(vertices, normals, faces, displacement_map, texture_map, dense_template):
    ''' Credit to Timo
    upsampling coarse mesh (with displacment map)
        vertices: vertices of coarse mesh, [nv, 3]
        normals: vertex normals, [nv, 3]
        faces: faces of coarse mesh, [nf, 3]
        texture_map: texture map, [256, 256, 3]
        displacement_map: displacment map, [256, 256]
        dense_template: 
    Returns: 
        dense_vertices: upsampled vertices with details, [number of dense vertices, 3]
        dense_colors: vertex color, [number of dense vertices, 3]
        dense_faces: [number of dense faces, 3]
    '''
    img_size = dense_template['img_size']
    dense_faces = dense_template['f']
    x_coords = dense_template['x_coords']
    y_coords = dense_template['y_coords']
    valid_pixel_ids = dense_template['valid_pixel_ids']
    valid_pixel_3d_faces = dense_template['valid_pixel_3d_faces']
    valid_pixel_b_coords = dense_template['valid_pixel_b_coords']

    pixel_3d_points = vertices[valid_pixel_3d_faces[:, 0], :] * valid_pixel_b_coords[:, 0][:, np.newaxis] + \
                    vertices[valid_pixel_3d_faces[:, 1], :] * valid_pixel_b_coords[:, 1][:, np.newaxis] + \
                    vertices[valid_pixel_3d_faces[:, 2], :] * valid_pixel_b_coords[:, 2][:, np.newaxis]
    vertex_normals = normals
    pixel_3d_normals = vertex_normals[valid_pixel_3d_faces[:, 0], :] * valid_pixel_b_coords[:, 0][:, np.newaxis] + \
                    vertex_normals[valid_pixel_3d_faces[:, 1], :] * valid_pixel_b_coords[:, 1][:, np.newaxis] + \
                    vertex_normals[valid_pixel_3d_faces[:, 2], :] * valid_pixel_b_coords[:, 2][:, np.newaxis]
    pixel_3d_normals = pixel_3d_normals / np.linalg.norm(pixel_3d_normals, axis=-1)[:, np.newaxis]
    displacements = displacement_map[y_coords[valid_pixel_ids].astype(int), x_coords[valid_pixel_ids].astype(int)]
    dense_colors = texture_map[y_coords[valid_pixel_ids].astype(int), x_coords[valid_pixel_ids].astype(int)]
    offsets = np.einsum('i,ij->ij', displacements, pixel_3d_normals)
    dense_vertices = pixel_3d_points + offsets
    return dense_vertices, dense_colors, dense_faces

# borrowed from https://github.com/YadiraF/PRNet/blob/master/utils/write.py
def write_obj(obj_name,
              vertices,
              faces,
              colors=None,
              texture=None,
              uvcoords=None,
              uvfaces=None,
              inverse_face_order=False,
              normal_map=None,
              ):
    ''' Save 3D face model with texture. 
    Ref: https://github.com/patrikhuber/eos/blob/bd00155ebae4b1a13b08bf5a991694d682abbada/include/eos/core/Mesh.hpp
    Args:
        obj_name: str
        vertices: shape = (nver, 3)
        colors: shape = (nver, 3)
        faces: shape = (ntri, 3)
        texture: shape = (uv_size, uv_size, 3)
        uvcoords: shape = (nver, 2) max value<=1
    '''
    if os.path.splitext(obj_name)[-1] != '.obj':
        obj_name = obj_name + '.obj'
    mtl_name = obj_name.replace('.obj', '.mtl')
    texture_name = obj_name.replace('.obj', '.png')
    material_name = 'FaceTexture'

    faces = faces.copy()
    # mesh lab start with 1, python/c++ start from 0
    faces += 1
    if inverse_face_order:
        faces = faces[:, [2, 1, 0]]
        if uvfaces is not None:
            uvfaces = uvfaces[:, [2, 1, 0]]

    # write obj
    with open(obj_name, 'w') as f:
        # first line: write mtlib(material library)
        # f.write('# %s\n' % os.path.basename(obj_name))
        # f.write('#\n')
        # f.write('\n')
        if texture is not None:
            f.write('mtllib %s\n\n' % os.path.basename(mtl_name))

        # write vertices
        if colors is None:
            for i in range(vertices.shape[0]):
                f.write('v {} {} {}\n'.format(vertices[i, 0], vertices[i, 1], vertices[i, 2]))
        else:
            for i in range(vertices.shape[0]):
                f.write('v {} {} {} {} {} {}\n'.format(vertices[i, 0], vertices[i, 1], vertices[i, 2], colors[i, 0], colors[i, 1], colors[i, 2]))

        # write uv coords
        if texture is None:
            for i in range(faces.shape[0]):
                f.write('f {} {} {}\n'.format(faces[i, 2], faces[i, 1], faces[i, 0]))
        else:
            for i in range(uvcoords.shape[0]):
                f.write('vt {} {}\n'.format(uvcoords[i,0], uvcoords[i,1]))
            f.write('usemtl %s\n' % material_name)
            # write f: ver ind/ uv ind
            uvfaces = uvfaces + 1
            for i in range(faces.shape[0]):
                f.write('f {}/{} {}/{} {}/{}\n'.format(
                    #  faces[i, 2], uvfaces[i, 2],
                    #  faces[i, 1], uvfaces[i, 1],
                    #  faces[i, 0], uvfaces[i, 0]
                    faces[i, 0], uvfaces[i, 0],
                    faces[i, 1], uvfaces[i, 1],
                    faces[i, 2], uvfaces[i, 2]
                )
                )
            # write mtl
            with open(mtl_name, 'w') as f:
                f.write('newmtl %s\n' % material_name)
                s = 'map_Kd {}\n'.format(os.path.basename(texture_name)) # map to image
                f.write(s)

                if normal_map is not None:
                    name, _ = os.path.splitext(obj_name)
                    normal_name = f'{name}_normals.png'
                    f.write(f'disp {normal_name}')
                    # out_normal_map = normal_map / (np.linalg.norm(
                    #     normal_map, axis=-1, keepdims=True) + 1e-9)
                    # out_normal_map = (out_normal_map + 1) * 0.5

                    cv2.imwrite(
                        normal_name,
                        # (out_normal_map * 255).astype(np.uint8)[:, :, ::-1]
                        normal_map
                    )
            cv2.imwrite(texture_name, texture)


## load obj,  similar to load_obj from pytorch3d
def load_obj(obj_filename):
    """ Ref: https://github.com/facebookresearch/pytorch3d/blob/25c065e9dafa90163e7cec873dbb324a637c68b7/pytorch3d/io/obj_io.py
    Load a mesh from a file-like object.
    """
    with open(obj_filename, 'r') as f:
        lines = [line.strip() for line in f]

    verts, uvcoords = [], []
    faces, uv_faces = [], []
    # startswith expects each line to be a string. If the file is read in as
    # bytes then first decode to strings.
    if lines and isinstance(lines[0], bytes):
        lines = [el.decode("utf-8") for el in lines]

    for line in lines:
        tokens = line.strip().split()
        if line.startswith("v "):  # Line is a vertex.
            vert = [float(x) for x in tokens[1:4]]
            if len(vert) != 3:
                msg = "Vertex %s does not have 3 values. Line: %s"
                raise ValueError(msg % (str(vert), str(line)))
            verts.append(vert)
        elif line.startswith("vt "):  # Line is a texture.
            tx = [float(x) for x in tokens[1:3]]
            if len(tx) != 2:
                raise ValueError(
                    "Texture %s does not have 2 values. Line: %s" % (str(tx), str(line))
                )
            uvcoords.append(tx)
        elif line.startswith("f "):  # Line is a face.
            # Update face properties info.
            face = tokens[1:]
            face_list = [f.split("/") for f in face]
            for vert_props in face_list:
                # Vertex index.
                faces.append(int(vert_props[0]))
                if len(vert_props) > 1:
                    if vert_props[1] != "":
                        # Texture index is present e.g. f 4/1/1.
                        uv_faces.append(int(vert_props[1]))

    verts = torch.tensor(verts, dtype=torch.float32)
    uvcoords = torch.tensor(uvcoords, dtype=torch.float32)
    faces = torch.tensor(faces, dtype=torch.long); faces = faces.reshape(-1, 3) - 1
    uv_faces = torch.tensor(uv_faces, dtype=torch.long); uv_faces = uv_faces.reshape(-1, 3) - 1
    return (
        verts,
        uvcoords,
        faces,
        uv_faces
    )

# ---------------------------- process/generate vertices, normals, faces
def generate_triangles(h, w, margin_x=2, margin_y=5, mask = None):
    # quad layout:
    # 0 1 ... w-1
    # w w+1
    #.
    # w*h
    triangles = []
    for x in range(margin_x, w-1-margin_x):
        for y in range(margin_y, h-1-margin_y):
            triangle0 = [y*w + x, y*w + x + 1, (y+1)*w + x]
            triangle1 = [y*w + x + 1, (y+1)*w + x + 1, (y+1)*w + x]
            triangles.append(triangle0)
            triangles.append(triangle1)
    triangles = np.array(triangles)
    triangles = triangles[:,[0,2,1]]
    return triangles

# borrowed from https://github.com/daniilidis-group/neural_renderer/blob/master/neural_renderer/vertices_to_faces.py
def face_vertices(vertices, faces):
    """ 
    :param vertices: [batch size, number of vertices, 3]
    :param faces: [batch size, number of faces, 3]
    :return: [batch size, number of faces, 3, 3]
    """
    assert (vertices.ndimension() == 3)
    assert (faces.ndimension() == 3)
    assert (vertices.shape[0] == faces.shape[0])
    assert (vertices.shape[2] == 3)
    assert (faces.shape[2] == 3)

    bs, nv = vertices.shape[:2]
    bs, nf = faces.shape[:2]
    device = vertices.device
    faces = faces + (torch.arange(bs, dtype=torch.int32).to(device) * nv)[:, None, None]
    vertices = vertices.reshape((bs * nv, 3))
    # pytorch only supports long and byte tensors for indexing
    return vertices[faces.long()]
    
def vertex_normals(vertices, faces):
    """
    :param vertices: [batch size, number of vertices, 3]
    :param faces: [batch size, number of faces, 3]
    :return: [batch size, number of vertices, 3]
    """
    assert (vertices.ndimension() == 3)
    assert (faces.ndimension() == 3)
    assert (vertices.shape[0] == faces.shape[0])
    assert (vertices.shape[2] == 3)
    assert (faces.shape[2] == 3)
    bs, nv = vertices.shape[:2]
    bs, nf = faces.shape[:2]
    device = vertices.device
    normals = torch.zeros(bs * nv, 3).to(device)

    faces = faces + (torch.arange(bs, dtype=torch.int32).to(device) * nv)[:, None, None] # expanded faces
    vertices_faces = vertices.reshape((bs * nv, 3))[faces.long()]

    faces = faces.reshape(-1, 3)
    vertices_faces = vertices_faces.reshape(-1, 3, 3)

    normals.index_add_(0, faces[:, 1].long(), 
                       torch.cross(vertices_faces[:, 2] - vertices_faces[:, 1], vertices_faces[:, 0] - vertices_faces[:, 1]))
    normals.index_add_(0, faces[:, 2].long(), 
                       torch.cross(vertices_faces[:, 0] - vertices_faces[:, 2], vertices_faces[:, 1] - vertices_faces[:, 2]))
    normals.index_add_(0, faces[:, 0].long(),
                       torch.cross(vertices_faces[:, 1] - vertices_faces[:, 0], vertices_faces[:, 2] - vertices_faces[:, 0]))

    normals = F.normalize(normals, eps=1e-6, dim=1)
    normals = normals.reshape((bs, nv, 3))
    # pytorch only supports long and byte tensors for indexing
    return normals

def batch_orth_proj(X, camera):
    ''' orthgraphic projection
        X:  3d vertices, [bz, n_point, 3]
        camera: scale and translation, [bz, 3], [scale, tx, ty]
    '''
    camera = camera.clone().view(-1, 1, 3)
    X_trans = X[:, :, :2] + camera[:, :, 1:]
    X_trans = torch.cat([X_trans, X[:,:,2:]], 2)
    shape = X_trans.shape
    Xn = (camera[:, :, 0:1] * X_trans)
    return Xn
    
# -------------------------------------- image processing
# borrowed from: https://torchgeometry.readthedocs.io/en/latest/_modules/kornia/filters
def gaussian(window_size, sigma):
    def gauss_fcn(x):
        return -(x - window_size // 2)**2 / float(2 * sigma**2)
    gauss = torch.stack(
        [torch.exp(torch.tensor(gauss_fcn(x))) for x in range(window_size)])
    return gauss / gauss.sum()

def get_gaussian_kernel(kernel_size: int, sigma: float):
    r"""Function that returns Gaussian filter coefficients.

    Args:
        kernel_size (int): filter size. It should be odd and positive.
        sigma (float): gaussian standard deviation.

    Returns:
        Tensor: 1D tensor with gaussian filter coefficients.

    Shape:
        - Output: :math:`(\text{kernel_size})`

    Examples::

        >>> kornia.image.get_gaussian_kernel(3, 2.5)
        tensor([0.3243, 0.3513, 0.3243])

        >>> kornia.image.get_gaussian_kernel(5, 1.5)
        tensor([0.1201, 0.2339, 0.2921, 0.2339, 0.1201])
    """
    if not isinstance(kernel_size, int) or kernel_size % 2 == 0 or \
            kernel_size <= 0:
        raise TypeError("kernel_size must be an odd positive integer. "
                        "Got {}".format(kernel_size))
    window_1d = gaussian(kernel_size, sigma)
    return window_1d

def get_gaussian_kernel2d(kernel_size, sigma):
    r"""Function that returns Gaussian filter matrix coefficients.

    Args:
        kernel_size (Tuple[int, int]): filter sizes in the x and y direction.
         Sizes should be odd and positive.
        sigma (Tuple[int, int]): gaussian standard deviation in the x and y
         direction.

    Returns:
        Tensor: 2D tensor with gaussian filter matrix coefficients.

    Shape:
        - Output: :math:`(\text{kernel_size}_x, \text{kernel_size}_y)`

    Examples::

        >>> kornia.image.get_gaussian_kernel2d((3, 3), (1.5, 1.5))
        tensor([[0.0947, 0.1183, 0.0947],
                [0.1183, 0.1478, 0.1183],
                [0.0947, 0.1183, 0.0947]])

        >>> kornia.image.get_gaussian_kernel2d((3, 5), (1.5, 1.5))
        tensor([[0.0370, 0.0720, 0.0899, 0.0720, 0.0370],
                [0.0462, 0.0899, 0.1123, 0.0899, 0.0462],
                [0.0370, 0.0720, 0.0899, 0.0720, 0.0370]])
    """
    if not isinstance(kernel_size, tuple) or len(kernel_size) != 2:
        raise TypeError("kernel_size must be a tuple of length two. Got {}"
                        .format(kernel_size))
    if not isinstance(sigma, tuple) or len(sigma) != 2:
        raise TypeError("sigma must be a tuple of length two. Got {}"
                        .format(sigma))
    ksize_x, ksize_y = kernel_size
    sigma_x, sigma_y = sigma
    kernel_x = get_gaussian_kernel(ksize_x, sigma_x)
    kernel_y = get_gaussian_kernel(ksize_y, sigma_y)
    kernel_2d = torch.matmul(
        kernel_x.unsqueeze(-1), kernel_y.unsqueeze(-1).t())
    return kernel_2d

def gaussian_blur(x, kernel_size=(3,3), sigma=(0.8,0.8)):
    b, c, h, w = x.shape
    kernel = get_gaussian_kernel2d(kernel_size, sigma).to(x.device).to(x.dtype)
    kernel = kernel.repeat(c, 1, 1, 1)
    padding = [(k - 1) // 2 for k in kernel_size]
    return F.conv2d(x, kernel, padding=padding, stride=1, groups=c)

def _compute_binary_kernel(window_size):
    r"""Creates a binary kernel to extract the patches. If the window size
    is HxW will create a (H*W)xHxW kernel.
    """
    window_range = window_size[0] * window_size[1]
    kernel: torch.Tensor = torch.zeros(window_range, window_range)
    for i in range(window_range):
        kernel[i, i] += 1.0
    return kernel.view(window_range, 1, window_size[0], window_size[1])

def median_blur(x, kernel_size=(3,3)):
    b, c, h, w = x.shape
    kernel = _compute_binary_kernel(kernel_size).to(x.device).to(x.dtype)
    kernel = kernel.repeat(c, 1, 1, 1)
    padding = [(k - 1) // 2 for k in kernel_size]
    features = F.conv2d(x, kernel, padding=padding, stride=1, groups=c)
    features = features.view(b,c,-1,h,w)
    median = torch.median(features, dim=2)[0]
    return median

def get_laplacian_kernel2d(kernel_size: int):
    r"""Function that returns Gaussian filter matrix coefficients.

    Args:
        kernel_size (int): filter size should be odd.

    Returns:
        Tensor: 2D tensor with laplacian filter matrix coefficients.

    Shape:
        - Output: :math:`(\text{kernel_size}_x, \text{kernel_size}_y)`

    Examples::

        >>> kornia.image.get_laplacian_kernel2d(3)
        tensor([[ 1.,  1.,  1.],
                [ 1., -8.,  1.],
                [ 1.,  1.,  1.]])

        >>> kornia.image.get_laplacian_kernel2d(5)
        tensor([[  1.,   1.,   1.,   1.,   1.],
                [  1.,   1.,   1.,   1.,   1.],
                [  1.,   1., -24.,   1.,   1.],
                [  1.,   1.,   1.,   1.,   1.],
                [  1.,   1.,   1.,   1.,   1.]])

    """
    if not isinstance(kernel_size, int) or kernel_size % 2 == 0 or \
            kernel_size <= 0:
        raise TypeError("ksize must be an odd positive integer. Got {}"
                        .format(kernel_size))

    kernel = torch.ones((kernel_size, kernel_size))
    mid = kernel_size // 2
    kernel[mid, mid] = 1 - kernel_size ** 2
    kernel_2d: torch.Tensor = kernel
    return kernel_2d

def laplacian(x):
    # https://torchgeometry.readthedocs.io/en/latest/_modules/kornia/filters/laplacian.html
    b, c, h, w = x.shape
    kernel_size = 3
    kernel = get_laplacian_kernel2d(kernel_size).to(x.device).to(x.dtype)
    kernel = kernel.repeat(c, 1, 1, 1)
    padding = (kernel_size - 1) // 2
    return F.conv2d(x, kernel, padding=padding, stride=1, groups=c)

def angle2matrix(angles):
    ''' get rotation matrix from three rotation angles(degree). right-handed.
    Args:
        angles: [batch_size, 3] tensor containing X, Y, and Z angles.
        x: pitch. positive for looking down.
        y: yaw. positive for looking left. 
        z: roll. positive for tilting head right. 
    Returns:
        R: [batch_size, 3, 3]. rotation matrices.
    '''
    angles = angles*(np.pi)/180.
    s = torch.sin(angles)
    c = torch.cos(angles)

    cx, cy, cz = (c[:, 0], c[:, 1], c[:, 2])
    sx, sy, sz = (s[:, 0], s[:, 1], s[:, 2])

    zeros = torch.zeros_like(s[:, 0]).to(angles.device)
    ones = torch.ones_like(s[:, 0]).to(angles.device)

    # Rz.dot(Ry.dot(Rx))
    R_flattened = torch.stack(
    [
      cz * cy, cz * sy * sx - sz * cx, cz * sy * cx + sz * sx,
      sz * cy, sz * sy * sx + cz * cx, sz * sy * cx - cz * sx,
          -sy,                cy * sx,                cy * cx,
    ],
    dim=0) #[batch_size, 9]
    R = torch.reshape(R_flattened, (-1, 3, 3)) #[batch_size, 3, 3]
    return R

def binary_erosion(tensor, kernel_size=5):
    # tensor: [bz, 1, h, w]. 
    device = tensor.device
    mask = tensor.cpu().numpy()
    structure=np.ones((kernel_size,kernel_size))
    new_mask = mask.copy()
    for i in range(mask.shape[0]):
        new_mask[i,0] = morphology.binary_erosion(mask[i,0], structure)
    return torch.from_numpy(new_mask.astype(np.float32)).to(device)

def flip_image(src_image, kps):
    '''
        purpose:
            flip a image given by src_image and the 2d keypoints
        flip_mode: 
            0: horizontal flip
            >0: vertical flip
            <0: horizontal & vertical flip
    '''
    h, w = src_image.shape[0], src_image.shape[1]
    src_image = cv2.flip(src_image, 1)
    if kps is not None:
        kps[:, 0] = w - 1 - kps[:, 0]
        kp_map = [5, 4, 3, 2, 1, 0, 11, 10, 9, 8, 7, 6, 12, 13]
        kps[:, :] = kps[kp_map]

    return src_image, kps

# -------------------------------------- io
def copy_state_dict(cur_state_dict, pre_state_dict, prefix='', load_name=None):
    def _get_params(key):
        key = prefix + key
        if key in pre_state_dict:
            return pre_state_dict[key]
        return None
    for k in cur_state_dict.keys():
        if load_name is not None:
            if load_name not in k:
                continue
        v = _get_params(k)
        try:
            if v is None:
                # print('parameter {} not found'.format(k))
                continue
            cur_state_dict[k].copy_(v)
        except:
            # print('copy param {} failed'.format(k))
            continue

def check_mkdir(path):
    if not os.path.exists(path):
        print('creating %s' % path)
        os.makedirs(path)

def check_mkdirlist(pathlist):
    for path in pathlist:
        if not os.path.exists(path):
            print('creating %s' % path)
            os.makedirs(path)

def tensor2image(tensor):
    image = tensor.detach().cpu().numpy()
    image = image*255.
    image = np.maximum(np.minimum(image, 255), 0)
    image = image.transpose(1,2,0)[:,:,[2,1,0]]
    return image.astype(np.uint8).copy()

def dict2obj(d):
    # if isinstance(d, list):
    #     d = [dict2obj(x) for x in d]
    if not isinstance(d, dict):
        return d
    class C(object):
        pass
    o = C()
    for k in d:
        o.__dict__[k] = dict2obj(d[k])
    return o

class Struct(object):
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)

# original saved file with DataParallel
def remove_module(state_dict):
# create new OrderedDict that does not contain `module.`
    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        name = k[7:] # remove `module.`
        new_state_dict[name] = v
    return new_state_dict

def dict_tensor2npy(tensor_dict):
    npy_dict = {}
    for key in tensor_dict:
        npy_dict[key] = tensor_dict[key][0].cpu().numpy()
    return npy_dict
        
# ---------------------------------- visualization
end_list = np.array([17, 21], dtype = np.int32) - 1
# end_list = np.array([17, 22, 27, 42, 48, 31, 36, 68], dtype = np.int32) - 1
def plot_kpts(image, kpts, color = 'r'):
    ''' Draw 68 key points
    Args: 
        image: the input image
        kpt: (68, 3).
    '''
    if color == 'r':
        c = (255, 0, 0)
    elif color == 'g':
        c = (0, 255, 0)
    elif color == 'b':
        c = (255, 0, 0)
    image = image.copy()
    kpts = kpts.copy()
    kpts = np.concatenate((kpts[:17],kpts[27:36]), axis=0)
    radius = max(int(min(image.shape[0], image.shape[1])/200), 1)
    for i in range(kpts.shape[0]-1):
        st = kpts[i, :2]
        if kpts.shape[1]==4:
            if kpts[i, 3] > 0.5:
                c = (0, 255, 0)
            else:
                c = (0, 0, 255)
        if i in end_list:
            continue
        ed = kpts[i + 1, :2]
        image = cv2.line(image, (int(st[0]), int(st[1])), (int(ed[0]), int(ed[1])), (255, 255, 255), radius)
        # image = cv2.circle(image,(int(st[0]), int(st[1])), radius, c, radius*2)  

    return image

def plot_verts(image, kpts, color = 'r'):
    ''' Draw 68 key points
    Args: 
        image: the input image
        kpt: (68, 3).
    '''
    if color == 'r':
        c = (255, 0, 0)
    elif color == 'g':
        c = (0, 255, 0)
    elif color == 'b':
        c = (0, 0, 255)
    elif color == 'y':
        c = (0, 255, 255)
    image = image.copy()

    for i in range(kpts.shape[0]):
        st = kpts[i, :2]
        image = cv2.circle(image,(int(st[0]), int(st[1])), 1, c, 2)  

    return image

def tensor_vis_landmarks(images, landmarks, gt_landmarks=None, color = 'g', isScale=True):
    # visualize landmarks
    vis_landmarks = []
    images = images.cpu().numpy()
    predicted_landmarks = landmarks.detach().cpu().numpy()
    if gt_landmarks is not None:
        gt_landmarks_np = gt_landmarks.detach().cpu().numpy()
    for i in range(images.shape[0]):
        image = images[i]
        image = image.transpose(1,2,0)[:,:,[2,1,0]].copy(); image = (image*255)
        predicted_landmark = predicted_landmarks[i]
        if predicted_landmark.shape[0] == 68:
            image_landmarks = plot_kpts(np.ones_like(image) * 127.5, predicted_landmark, color)
            if gt_landmarks is not None:
                image_landmarks = plot_verts(image_landmarks, gt_landmarks_np[i]*image.shape[0]/2 + image.shape[0]/2, 'r')
        else:
            image_landmarks = plot_verts(image, predicted_landmark, color)
            if gt_landmarks is not None:
                image_landmarks = plot_verts(image_landmarks, gt_landmarks_np[i]*image.shape[0]/2 + image.shape[0]/2, 'r')
        vis_landmarks.append(image_landmarks)
    vis_landmarks = np.stack(vis_landmarks)
    vis_landmarks = torch.from_numpy(vis_landmarks[:,:,:,[2,1,0]].transpose(0,3,1,2))/255.#, dtype=torch.float32)
    return vis_landmarks


############### for training
def load_local_mask(image_size=256, mode='bbx'):
    if mode == 'bbx':
        # UV space face attributes bbx in size 2048 (l r t b)
        # face = np.array([512, 1536, 512, 1536]) #
        face = np.array([400, 1648, 400, 1648])
        # if image_size == 512:
            # face = np.array([400, 400+512*2, 400, 400+512*2])
            # face = np.array([512, 512+512*2, 512, 512+512*2])

        forehead = np.array([550, 1498, 430, 700+50])
        eye_nose = np.array([490, 1558, 700, 1050+50])
        mouth = np.array([574, 1474, 1050, 1550])
        ratio = image_size / 2048.
        face = (face * ratio).astype(np.int)
        forehead = (forehead * ratio).astype(np.int)
        eye_nose = (eye_nose * ratio).astype(np.int)
        mouth = (mouth * ratio).astype(np.int)
        regional_mask = np.array([face, forehead, eye_nose, mouth])

    return regional_mask

def visualize_grid(visdict, savepath=None, size=224, dim=1, return_gird=True):
    '''
    image range should be [0,1]
    dim: 2 for horizontal. 1 for vertical
    '''
    assert dim == 1 or dim==2
    grids = {}
    for key in visdict:
        _,_,h,w = visdict[key].shape
        if dim == 2:
            new_h = size; new_w = int(w*size/h)
        elif dim == 1:
            new_h = int(h*size/w); new_w = size
        grids[key] = torchvision.utils.make_grid(F.interpolate(visdict[key], [new_h, new_w]).detach().cpu())
    grid = torch.cat(list(grids.values()), dim)
    grid_image = (grid.numpy().transpose(1,2,0).copy()*255)[:,:,[2,1,0]]
    grid_image = np.minimum(np.maximum(grid_image, 0), 255).astype(np.uint8)
    if savepath:
        cv2.imwrite(savepath, grid_image)
    if return_gird:
        return grid_image


import sys
sys.path.append('./MyModel')

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from .sub_nets import VGG19_pytorch, Blend_Net, Warp_Net

class MyGenerator(nn.Module):
    def __init__(self):
        super(MyGenerator, self).__init__()
        self.vgg19 = VGG19_pytorch()
        self.warp_net = Warp_Net()
        self.blend_net = Blend_Net(5)
        self.net_dict = {}
        
    # source color + target structure
    def forward(self, source_rgb, target_rgb, source_onehot, target_onehot, target_gray, target_face_mask):
    # def forward(self, target_gray, target_onehot, source_rgb, source_onehot):
        self.net_dict['target_rgb'] = target_rgb
        self.net_dict['target_gray'] = target_gray
        self.net_dict['target_onehot'] = target_onehot
        self.net_dict['source_rgb'] = source_rgb
        self.net_dict['source_onehot'] = source_onehot
        self.net_dict['target_face_mask'] = target_face_mask
        
        """
        Step 1 : get features
        input 0 ~ 1
        """
        with torch.no_grad():
            self.net_dict["target_rgb_features"] : list = self.vgg19(
                F.interpolate((self.net_dict["target_rgb"]+1)/2, (256,256)), ["r12", "r22", "r32", "r42", "r52"], preprocess=True
            )
            
            self.net_dict["source_rgb_features"] : list = self.vgg19(
                F.interpolate((self.net_dict["source_rgb"]+1)/2, (256,256)), ["r12", "r22", "r32", "r42", "r52"], preprocess=True
            )
            
        # "r12", "r22", "r32", "r42", "r52"
        for i, (target_feature, source_feature) in enumerate(zip(self.net_dict["target_rgb_features"], self.net_dict["source_rgb_features"])):
            self.net_dict["target_rgb_features"][i] = self._feature_normalize(target_feature)
            self.net_dict["source_rgb_features"][i] = self._feature_normalize(source_feature)
        
        """
        Step 2 : get color reference map & similarity map
        vgg relu results(source & target) -> [ Warp Net ] -> resized feature maps(source & target)(img size / 4)
        output : color reference map(color ab), similarity map(b,1,h/4,h/4)?
        """
        self.net_dict["target_rgb_features_small"], self.net_dict["source_rgb_features_small"] = self.warp_net(
            self.net_dict["target_rgb_features"][1],
            self.net_dict["target_rgb_features"][2],
            self.net_dict["target_rgb_features"][3],
            self.net_dict["target_rgb_features"][4],
            self.net_dict["source_rgb_features"][1],
            self.net_dict["source_rgb_features"][2],
            self.net_dict["source_rgb_features"][3],
            self.net_dict["source_rgb_features"][4]
        )

        # color_reference_lab(with source l)
        # b 3 h 216 384 / b 1 h 218 384
        self.get_color_map()
        self.net_dict["color_reference_innerface_rgb"] = self.net_dict["color_reference_rgb"] * self.net_dict["target_face_mask"]
        blend_input = torch.cat((self.net_dict["target_gray"], self.net_dict["target_face_mask"], self.net_dict["color_reference_innerface_rgb"]), dim=1)
        self.net_dict["fake_rgb"] = self.blend_net(blend_input.clone().detach())
        return self.net_dict["fake_rgb"], self.net_dict["color_reference_rgb"]
        

    def get_color_map(self):

        _, n_ch, origin_h, origin_w = self.net_dict['target_onehot'].shape
        b, c, h, w = self.net_dict['target_rgb_features_small'].size()
        target_mask_quater = F.interpolate(self.net_dict['target_onehot'], size=(h, w), mode='nearest')
        source_mask_quater = F.interpolate(self.net_dict['source_onehot'], size=(h, w), mode='nearest')
        source_color_quater = F.interpolate(self.net_dict['source_rgb'], size=(h, w), mode='bilinear')
        target_color_quater = F.interpolate(self.net_dict['target_rgb'], size=(h, w), mode='bilinear')
        canvas = torch.zeros_like(source_color_quater)

        # target_mask_quater[:,-1] = torch.logical_or(target_mask_quater[:,0], target_mask_quater[:,-1])

        for b_idx in range(b):
            for c_idx in range(1, n_ch):
                target_1ch_mask, source_1ch_mask = target_mask_quater[b_idx,c_idx], source_mask_quater[b_idx,c_idx]

                # if target_1ch_mask.sum() != 0 and source_1ch_mask.sum() == 0:
                if target_1ch_mask.sum() != 0 and source_1ch_mask.sum() < 9 :
                    target_matrix = torch.masked_select(self.net_dict['target_rgb_features_small'][b_idx], target_1ch_mask.bool()).reshape(c, -1) # 64, pixel_num_A
                    target_matrix_bar = target_matrix - target_matrix.mean(1, keepdim=True) # 64, pixel_num_A
                    if target_1ch_mask.sum() == 1:
                        target_matrix_ = target_matrix_bar
                    else:
                        target_matrix_norm = torch.norm(target_matrix_bar, dim=0, keepdim=True)
                        target_matrix_ = target_matrix_bar / target_matrix_norm

                    source_matrix = torch.masked_select(self.net_dict['target_rgb_features_small'][b_idx], target_1ch_mask.bool()).reshape(c, -1) # 64, pixel_num_B
                    source_matrix_bar = source_matrix - source_matrix.mean(1, keepdim=True) # 64, pixel_num_B
                    if target_1ch_mask.sum() == 1:
                        source_matrix_ = source_matrix_bar
                    else:
                        source_matrix_norm = torch.norm(source_matrix_bar, dim=0, keepdim=True)
                        source_matrix_ = source_matrix_bar / source_matrix_norm
                

                    correlation_matrix = torch.matmul(target_matrix_.transpose(0,1), source_matrix_)
                    correlation_matrix = F.softmax(correlation_matrix,dim=1)

                    source_pixels = torch.masked_select(target_color_quater[b_idx], target_1ch_mask.bool()).reshape(3,-1) # 3
                    colorized_matrix = torch.matmul(correlation_matrix, source_pixels.transpose(0,1)).transpose(0,1)

                    canvas[b_idx].masked_scatter_(target_1ch_mask.bool(), colorized_matrix) # 3 128 128
                    
                    
                elif target_1ch_mask.sum() != 0:
                
                # if target_1ch_mask.sum() == 0 or target_1ch_mask.sum() == 1 or source_1ch_mask.sum() == 0 or source_1ch_mask.sum() == 1:
                #     continue
            
                    target_matrix = torch.masked_select(self.net_dict['target_rgb_features_small'][b_idx], target_1ch_mask.bool()).reshape(c, -1) # 64, pixel_num_A
                    target_matrix_bar = target_matrix - target_matrix.mean(1, keepdim=True) # 64, pixel_num_A
                    if target_1ch_mask.sum() == 1:
                        target_matrix_ = target_matrix_bar
                    else:
                        target_matrix_norm = torch.norm(target_matrix_bar, dim=0, keepdim=True)
                        target_matrix_ = target_matrix_bar / target_matrix_norm

                    source_matrix = torch.masked_select(self.net_dict['source_rgb_features_small'][b_idx], source_1ch_mask.bool()).reshape(c, -1) # 64, pixel_num_B
                    source_matrix_bar = source_matrix - source_matrix.mean(1, keepdim=True) # 64, pixel_num_B
                    if source_1ch_mask.sum() == 1:
                        source_matrix_ = source_matrix_bar
                    else:
                        source_matrix_norm = torch.norm(source_matrix_bar, dim=0, keepdim=True)
                        source_matrix_ = source_matrix_bar / source_matrix_norm
                
                    correlation_matrix = torch.matmul(target_matrix_.transpose(0,1), source_matrix_)
                    correlation_matrix = F.softmax(correlation_matrix,dim=1)

                    source_pixels = torch.masked_select(source_color_quater[b_idx], source_1ch_mask.bool()).reshape(3,-1) # 3
                    colorized_matrix = torch.matmul(correlation_matrix, source_pixels.transpose(0,1)).transpose(0,1)

                    canvas[b_idx].masked_scatter_(target_1ch_mask.bool(), colorized_matrix) # 3 128 128




        self.net_dict['color_reference_rgb'] = F.interpolate(canvas, size=(origin_h, origin_w), mode='bilinear').clip(-1,1)
        
    def _feature_normalize(self, feature_in):
        feature_in_norm = torch.norm(feature_in, 2, 1, keepdim=True) + sys.float_info.epsilon
        feature_in_norm = torch.div(feature_in, feature_in_norm)
        return feature_in_norm


import os
import torch
import torch.nn as nn

from .decoders import res_shortcut_decoder
from .encoders.MatteFormer import MatteFormer


def get_generator(is_train=True):
    generator = Generator_MatteFormer(is_train=is_train)
    return generator


class Generator_MatteFormer(nn.Module):

    def __init__(self, is_train=True, pretrained_path='ckpt/swin_tiny_patch4_window7_224.pth'):

        super(Generator_MatteFormer, self).__init__()
        self.encoder = MatteFormer(embed_dim=96,
                                   depths=[2,2,6,2], # tiny-model
                                   num_heads=[3,6,12,24],
                                   window_size=7,
                                   mlp_ratio=4.0,
                                   qkv_bias=True,
                                   qk_scale=None,
                                   drop_rate=0.0,
                                   attn_drop_rate=0.0,
                                   drop_path_rate=0.3,
                                   patch_norm=True,
                                   use_checkpoint=False
                                   )
        # original
        self.decoder = res_shortcut_decoder()
        # self.decoder = decoders.__dict__['res_shortcut_decoder']()

        # if is_train:
        #     self.init_pretrained_weight(pretrained_path=pretrained_path)  # MatteFormer

    def init_pretrained_weight(self, pretrained_path=None):
        if not os.path.isfile(pretrained_path):
            print('Please Check your Pretrained weight path.. file not exist : {}'.format(pretrained_path))
            exit()
        weight = torch.load(pretrained_path)['model']

        # [1] get backbone weights
        weight_ = {}
        for i, (k, v) in enumerate(weight.items()):
            head = k.split('.')[0]
            if head in ['patch_embed', 'layers']:
                if 'attn_mask' in k:
                    print('[{}/{}] {} will be ignored'.format(i, len(weight.items()), k))
                    continue
                weight_.update({k: v})
            else:
                print('[{}/{}] {} will be ignored'.format(i, len(weight.items()), k))

        patch_embed_weight = weight_['patch_embed.proj.weight']
        patch_embed_weight_new = torch.nn.init.xavier_normal_(torch.randn(96, (3 + 3), 4, 4).cuda())
        patch_embed_weight_new[:, :3, :, :].copy_(patch_embed_weight)
        weight_['patch_embed.proj.weight'] = patch_embed_weight_new

        attn_layers = [k for k, v in weight_.items() if 'attn.relative_position_bias_table' in k]
        for layer_name in attn_layers:
            pos_bias = weight_[layer_name]
            n_bias, n_head = pos_bias.shape

            layer_idx, block_idx = int(layer_name.split('.')[1]), int(layer_name.split('.')[3])
            n_prior = block_idx + 1
            pos_bias_new = torch.nn.init.xavier_normal_(torch.randn(n_bias + n_prior*3, n_head))

            pos_bias_new[:n_bias, :] = pos_bias
            weight_[layer_name] = pos_bias_new

        attn_layers = [k for k, v in weight_.items() if 'attn.relative_position_index' in k]
        for layer_name in attn_layers:
            pos_index = weight_[layer_name]

            layer_idx, block_idx = int(layer_name.split('.')[1]), int(layer_name.split('.')[3])
            n_prior = block_idx + 1

            num_patch = 49
            last_idx = 169
            pos_index_new = torch.ones((num_patch, num_patch + n_prior*3)).long() * last_idx
            pos_index_new[:num_patch, :num_patch] = pos_index
            for i in range(n_prior):
                for j in range(3):
                    pos_index_new[:, num_patch + i*3 + j:num_patch + i*3 +j +1] = last_idx + i*3 + j
            weight_[layer_name] = pos_index_new

        self.encoder.load_state_dict(weight_, strict=False)
        print('load pretrained model done')

    def forward(self, image, trimap):
        inp = torch.cat((image, trimap), axis=1)
        x = self.encoder(inp, trimap)
        embedding = x[-1]
        outs = self.decoder(embedding, x[:-1])
        return outs


if __name__ == '__main__':

    img = torch.ones([2, 3, 512, 512]).cuda()
    tri = torch.ones([2, 3, 512, 512]).cuda()

    generator = Generator_MatteFormer().cuda()

    inp1 = torch.Tensor(2, 3, 512, 512).cuda()
    inp2 = torch.ones(2, 3, 512, 512).cuda()
    out = generator(inp1, inp2)

    print('Done')



import os

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms
from ..utils import check_ckpt_exist, convert_image_type, get_url_id
from .nets import MyGenerator

class DeBlurrer(nn.Module):
    def __init__(self, size=1024, folder_name='deblurrer', ckpt_name='G_1024_65000.pt', force=False, device='cuda'):
        """
        Input
        ---------
            - dtype : tensor
            - shape : (1, 3, 1024, 1024)
            - min max : (-1, 1)
            
            
        Output
        ---------
            fake
                - dtype : tensor
                - shape : (1, 3, 1024, 1024)
                - min max : (-1, 1)
            fake_res
                - dtype : tensor
                - shape : (1, 3, 1024, 1024)
                - min max : (-1, 1)
            edge
                - dtype : tensor
                - shape : (1, 3, 1024, 1024)
                - min max : (-1, 1)
        """
        super(DeBlurrer, self).__init__()
        self.device = device
        self.size = size
        url_id = get_url_id('~/.invz_pack/', folder_name, ckpt_name)
        root = os.path.join('~/.invz_pack/', folder_name)
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        ckpt = torch.load(ckpt_path, map_location=self.device)
        
        self.deblurrer = MyGenerator().to(self.device)
        self.deblurrer.load_state_dict(ckpt['model'])
        for param in self.deblurrer.parameters():
            param.requires_grad = False
        self.deblurrer.eval()
        del ckpt
        
    def forward(self, tensor_image, output_size=None, dicts={}):
        _, _, origin_h, origin_w = tensor_image.shape
        tensor_image = F.interpolate(tensor_image, (self.size, self.size), mode='bilinear')
        fake, fake_res, edge = self.deblurrer(tensor_image)
        if output_size == None:
            _fake = F.interpolate(fake, (origin_h, origin_w), mode='bilinear')
            _fake_res = F.interpolate(fake_res, (origin_h, origin_w), mode='bilinear')
            _edge = F.interpolate(edge, (origin_h, origin_w), mode='bilinear')
        else:
            _fake = F.interpolate(fake, (output_size, output_size), mode='bilinear')
            _fake_res = F.interpolate(fake_res, (output_size, output_size), mode='bilinear')
            _edge = F.interpolate(edge, (output_size, output_size), mode='bilinear')
        
        dicts['image'] = tensor_image
        dicts['fake'] = _fake
        dicts['fake_res'] = _fake_res
        dicts['edge'] = _edge
        return dicts



import cv2
from innerverz import Data_Process

image_path = 'sample.png'
DP = Data_Process()

"""
        face alignment usage
"""
from innerverz import FaceAligner
FA = FaceAligner(align_style='ffhq')
cv2_image = cv2.imread(image_path)
aligned_face, tfm, tfm_inv, lms_5, lms_106, face_bool = FA.get_face(cv2_image)
face_bool, lms_106, lms_5 = FA.detect_lmk(cv2_image)
aligned_face, tfm, tfm_inv = FA.align_face_from_5(cv2_image, lms_5)


"""
        Deep3dmm usage
"""
from innerverz import Deep3DMM
D3 = Deep3DMM()

ts_256_aligned_face = DP.image_pp(aligned_face, 256, normalize=True)
coeffs, lms_68 = D3(ts_256_aligned_face)


import os
import sys

import torch
import torch.nn as nn

from .net import MyGenerator
from ..utils import check_ckpt_exist, get_url_id

class Reage(nn.Module):
    def __init__(self, folder_name='reage', ckpt_name = 'G_512_130k.pt', force=False, device = 'cuda'):
        """
        Related Links
        --------
        https://studios.disneyresearch.com/2022/11/30/production-ready-face-re-aging-for-visual-effects/
        
        Options
        --------
        ckpt_name
            - G_512_130k.pt : image size(512)
            - G_1024_110k.pt : image size(1024)
        
        Forwards
        --------
        input
            - face image
                - aligned face with ffhq rules
                - dtype : tensor
                - shape : (b, 3, h, w)
                - min max : (-1 1)
            
            - mask 
                - skin(1) + brow(2,3) + ears(7,8) + nose(10) + mouth(11) + lip(12,13) + neck(14) area
                - dtype : tensor
                - shape : (b, 1, h, w)
                - min max : (0 or 1)
            - from_age, to_age
                - dtype : int
                - min max : (20 ~ 80)
            
        output
            dicts
            'result'
                - dtype : tensor
                - shape : (b, 3, h, w)
                - min max : (-1 1)
            'delta'
                - dtype : tensor
                - shape : (b, 3, h, w)
                - min max : (-1 1)
            
        """
        super(Reage, self).__init__()
        self.device = device
        self.generator = MyGenerator().to(self.device)
        
        url_id = get_url_id('~/.invz_pack/', folder_name, ckpt_name)
        root = os.path.join('~/.invz_pack/', folder_name)
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        ckpt = torch.load(ckpt_path, map_location=self.device)

        self.generator.load_state_dict(ckpt['model'])
        for param in self.generator.parameters():
            param.requires_grad = False
        self.generator.eval()
        del ckpt
        
    def forward(self, masked_image, mask, from_age, to_age, dicts={}) -> dict:
        
        model_input = torch.cat((masked_image, mask * from_age / 100, mask * to_age / 100), dim=1)
        
        delta = self.generator(model_input)
        result = masked_image + delta
        
        
        dicts['image'] = masked_image
        dicts['mask'] = mask
        dicts['from_age'] = from_age
        dicts['to_age'] = to_age
        dicts['result'] = result
        dicts['delta'] = delta
        return  dicts


import os

import torch
import torch.nn as nn
import torch.nn.functional as F

from .nets import BiSeNet
from ..utils import check_ckpt_exist, get_url_id, get_one_hot, arrange_mask

class FaceParser(nn.Module):
    def __init__(self, folder_name='face_parser', ckpt_name = 'faceparser.pth', force=False, device = 'cuda'):
        """
        Related Links
        --------
        https://github.com/zllrunning/face-parsing.PyTorch
        
        Label Index
        --------
        
        face_parts | bg| skin | Lbrow | Rbrow | Leye | Reye | glasses | Lear | Rear | ear_ring | nose | mouth | upper_lip | lower_lip | neck | neckless | cloth | hair | hat   
        --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | ---  
        label_index | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18  
        
        Methods
        ---------
        - get_label
        - get_onehot
        
        Returns
        ---------
        dict
            - 'label'
            - 'onehot'
        """
        super(FaceParser, self).__init__()

        self.device = device
        self.parsing_net = BiSeNet(n_classes=19).to(self.device)
        
        url_id = get_url_id('~/.invz_pack/', folder_name, ckpt_name)
        root = os.path.join('~/.invz_pack/', folder_name)
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        ckpt = torch.load(ckpt_path, map_location=self.device)
        
        self.parsing_net.load_state_dict(ckpt)
        for param in self.parsing_net.parameters():
            param.requires_grad = False
        self.parsing_net.eval()
        del ckpt
        
    def get_label(self, tensor_img, output_size=512, dicts={}) -> dict:
        """
        Input
        ---------
            - dtype : tensor
            - shape : (b, 3, 512, 512)
            - min max : (-1, 1)
            
        Output
        ---------
            - dtype : tensor
            - shape : (b, 512, 512)
            - min max : (0, 18)
        """
        label = self.parsing_net(tensor_img)
        _label = F.interpolate(label, (output_size,output_size), mode='bilinear').max(1)[1]
        _label = arrange_mask(_label, output_size)
        
        dicts['label'] = _label
        
        return dicts
    
    def get_onehot(self, tensor_img, size=512, dicts={}) -> dict:
        """
        Input
        ---------
            - dtype : tensor
            - shape : (b, 3, 512, 512)
            - min max : (-1, 1)
            
        Output
        ---------
            - dtype : tensor
            - shape : (b, 19, size, size)
            - min max : (0 or 1)
        """
        label = self.get_label(tensor_img, size)
        onehot = get_one_hot(label.unsqueeze(0))
        dicts['onehot'] = onehot
        return dicts


from .resnet_decoder import ResNet_D_Dec, ResShortCut_D_Decoder, BasicBlock

__all__ = ['res_shortcut_decoder']

def res_shortcut_decoder(**kwargs):
    return ResShortCut_D_Decoder(BasicBlock, [2, 3, 3, 2], **kwargs)




import os
import sys
sys.path.append('../')

import torch
import torch.nn as nn
import torch.nn.functional as F


from .nets import MyGenerator

from torchvision.transforms.functional import to_tensor, rgb_to_grayscale
from ..utils import check_ckpt_exist, get_url_id

class ReLighter(nn.Module):
    def __init__(self, folder_name='relighter', ckpt_name = 'G_129k.pt', force=False, device = 'cuda'):
        """
        Input
        --------
        - masked_gray_face
            - dtype : tensor
            - shape : (b, 1, 512, 512)
            - min max (-1 1)
            
        - relight_detail_shape
            - dtype : tensor
            - shape : (b, 1, 512, 512)
            - min max : (-1 1)
            
        - code_dict
            - dtype : dict
            
        - id_param
            - dtype : tensor
            - shape : (b 512)
        
        Output
        --------
        - result
            dicts
            'results'
                - dtype : tensor
                - shape : (b, 1 , 512, 512)
                - min max : (-1 1)
            'res'
                - dtype : tensor
                - shape : (b, 1 , 512, 512)
                - min max : (-1 1)
        """
        super(ReLighter, self).__init__()
        self.device = device
        
        self.img_size = 512
        self.generator = MyGenerator().to(self.device)

        url_id = get_url_id('~/.invz_pack/', folder_name, ckpt_name)
        root = os.path.join('~/.invz_pack/', folder_name)
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        
        ckpt = torch.load(os.path.join(ckpt_path), map_location=self.device)
        self.generator.load_state_dict(ckpt['model'])
        for param in self.generator.parameters():
            param.requires_grad = False
        self.generator.eval()
        del ckpt

    # source light -> target light
    def forward(self, masked_gray_face, relight_detail_shape, code_dict, id_param, output_size=None, dicts={}):
        _, _, origin_h, origin_w = masked_gray_face.shape
        
        masked_gray_face = F.interpolate(masked_gray_face, (self.img_size, self.img_size), mode='bilinear')
        relight_detail_shape = F.interpolate(relight_detail_shape, (self.img_size, self.img_size), mode='bilinear')
        
        params = torch.cat((code_dict['light'].view(1, -1), code_dict['cam'].view(1, -1), code_dict['pose'].view(1, -1), code_dict['detail'].view(1, -1), id_param), dim=-1)
        res = self.generator(masked_gray_face, relight_detail_shape, params)
        result = (res + masked_gray_face).clip(-1,1)
        
        if output_size == None:
            _result = F.interpolate(result, (origin_h, origin_w), mode='bilinear')
            _res = F.interpolate(res, (origin_h, origin_w), mode='bilinear')
        else:
            _result = F.interpolate(result, (output_size, output_size), mode='bilinear')
            _res = F.interpolate(res, (output_size, output_size), mode='bilinear')
        
        dicts['image'] = masked_gray_face
        dicts['detail_shape'] = relight_detail_shape
        dicts['result'] = _result
        dicts['res'] = _res
        return dicts
        


import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from .sub_nets.dense_motion import DenseMotionNetworkReg
from .sub_nets.LadderEncoder import LadderEncoder


class OcclusionAwareSPADEGenerator(nn.Module):
    """
    Generator that given source image, source ldmk image and driving ldmk image try to transform image according to movement trajectories
    according to the ldmks.
    """

    def __init__(
        self,
        block_expansion=32,
        max_features=1024,
        with_warp_im=False,
        hasid=False,
        with_gaze_htmap=False,
        with_ldmk_line=False,
        with_mouth_line=False,
        with_ht=False,
        use_IN=False,
        use_SN=True
    ):
        super(OcclusionAwareSPADEGenerator, self).__init__()
        self.with_warp_im = with_warp_im
        self.with_gaze_htmap = with_gaze_htmap
        self.with_ldmk_line = with_ldmk_line
        self.with_mouth_line = with_mouth_line
        self.with_ht = with_ht
        self.use_IN = use_IN
        self.use_SN = use_SN

        ##
        ladder_norm_type = "spectralinstance" if use_SN else "instance"
        self.ladder_network = LadderEncoder(
            need_feat=False,
            use_mask=False,
            label_nc=0,
            z_dim=512,
            norm_type=ladder_norm_type
        )
        self.dense_motion_network = DenseMotionNetworkReg(
            label_nc=0,
            ldmkimg=True,
            occlusion=True,
            block_expansion=64,
            max_features=1024,
            num_blocks=5,
            dec_lease=2,
            Lwarp=True,
            AdaINc=512
        )


        self.apply(self._init_weights)

    def _init_weights(self, m):
        gain = 0.02
        if isinstance(m, nn.LayerNorm):
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
            if m.weight is not None:
                nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.BatchNorm2d):
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
            if m.weight is not None:
                nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            if m.weight is not None:
                nn.init.xavier_normal_(m.weight, gain=gain)
            if hasattr(m, "bias") and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.Linear):
            if hasattr(m, "weight") and m.weight is not None:
                nn.init.xavier_normal_(m.weight, gain=gain)
            if hasattr(m, "bias") and m.bias is not None:
                nn.init.constant_(m.bias, 0)
                
    def forward(
            self,
            source_image,
            ldmk_line=None
        ):
        output_dict = {}

        input_t = (
            source_image
            if ldmk_line is None
            else torch.cat((source_image, ldmk_line), dim=1)
        )
        style_feat, _ = self.ladder_network(input_t)
        drv_exp = style_feat
        dense_motion = self.dense_motion_network(input_t, drv_exp)

        output_dict["deformation"] = dense_motion["deformation"]

        return output_dict, dense_motion

class MyGenerator(nn.Module):
    def __init__(self):
        super(MyGenerator, self).__init__()
        self.generator = OcclusionAwareSPADEGenerator(block_expansion=64, max_features=512, with_gaze_htmap=True, with_mouth_line=True, with_ldmk_line=True, use_IN=True, hasid=True)

    def forward(self, source_image, ldmk_line):
        output_dict, dense_motion = self.generator(source_image, ldmk_line)
        output_dict["deformed"] = self.deform_input(
            source_image, dense_motion["deformation"]
        )
        return output_dict["deformed"], dense_motion["deformation"]
    
    
    def deform_input(self, inp, deformation):
        _, h_old, w_old, _ = deformation.shape
        _, _, h, w = inp.shape
        if h_old != h or w_old != w:
            deformation = deformation.permute(0, 3, 1, 2)
            deformation = F.interpolate(deformation, size=(h, w), mode="bilinear")
            deformation = deformation.permute(0, 2, 3, 1)
        return F.grid_sample(
            inp.to(deformation.dtype), deformation, padding_mode="reflection"
        )


import torch
from   torch import nn
from   torch.nn import Parameter


def l2normalize(v, eps=1e-12):
    return v / (v.norm() + eps)


class SpectralNorm(nn.Module):
    """
    Based on https://github.com/heykeetae/Self-Attention-GAN/blob/master/spectral.py
    and add _noupdate_u_v() for evaluation
    """
    def __init__(self, module, name='weight', power_iterations=1):
        super(SpectralNorm, self).__init__()
        self.module = module
        self.name = name
        self.power_iterations = power_iterations
        if not self._made_params():
            self._make_params()

    def _update_u_v(self):
        u = getattr(self.module, self.name + "_u")
        v = getattr(self.module, self.name + "_v")
        w = getattr(self.module, self.name + "_bar")

        height = w.data.shape[0]
        for _ in range(self.power_iterations):
            v.data = l2normalize(torch.mv(torch.t(w.view(height,-1).data), u.data))
            u.data = l2normalize(torch.mv(w.view(height,-1).data, v.data))

        sigma = u.dot(w.view(height, -1).mv(v))
        setattr(self.module, self.name, w / sigma.expand_as(w))

    def _noupdate_u_v(self):
        u = getattr(self.module, self.name + "_u")
        v = getattr(self.module, self.name + "_v")
        w = getattr(self.module, self.name + "_bar")

        height = w.data.shape[0]
        sigma = u.dot(w.view(height, -1).mv(v))
        setattr(self.module, self.name, w / sigma.expand_as(w))

    def _made_params(self):
        try:
            u = getattr(self.module, self.name + "_u")
            v = getattr(self.module, self.name + "_v")
            w = getattr(self.module, self.name + "_bar")
            return True
        except AttributeError:
            return False

    def _make_params(self):
        w = getattr(self.module, self.name)

        height = w.data.shape[0]
        width = w.view(height, -1).data.shape[1]

        u = Parameter(w.data.new(height).normal_(0, 1), requires_grad=False)
        v = Parameter(w.data.new(width).normal_(0, 1), requires_grad=False)
        u.data = l2normalize(u.data)
        v.data = l2normalize(v.data)
        w_bar = Parameter(w.data)

        del self.module._parameters[self.name]

        self.module.register_parameter(self.name + "_u", u)
        self.module.register_parameter(self.name + "_v", v)
        self.module.register_parameter(self.name + "_bar", w_bar)

    def forward(self, *args):
        # if torch.is_grad_enabled() and self.module.training:
        if self.module.training:
            self._update_u_v()
        else:
            self._noupdate_u_v()
        return self.module.forward(*args)



from setuptools import setup, find_packages

requirements =[
    'Pillow',
    'numpy',
    'tqdm',
    'gdown',
    'insightface',
    'opencv-python',
    'timm',
    'kornia',
    'onnxruntime'
]

pypandoc_enabled = True
try:
    import pypandoc
    print('pandoc enabled')
    long_description = pypandoc.convert_file('README.md', 'rst')
except (IOError, ImportError, ModuleNotFoundError):
    print('WARNING: pandoc not enabled')
    long_description = open('README.md').read()
    pypandoc_enabled = False

setup(
    name="innerverz",
    version="0.1.11",
    author="Innerverz",
    author_email="study@innerverz.com",
    description="innerverz package",
    long_description=long_description,
    python_requires=">=3.6",
    install_requires=requirements,
    packages=find_packages()
    
    
)



import os, sys
# sys.path.append('./innerverz_package/video_fraceparser')

import cv2
from tqdm import tqdm
import torch
from torch import nn
import torch.nn.functional as F
from .utils.util import get_batch_labels, get_batch_frames
from .utils.util import InputPadder
from .utils.util import warp, get_list_chunk, arrange_mask
from .nets import RAFT

from ..utils import check_ckpt_exist, convert_image_type, get_url_id
from ..face_parser import FaceParser

class Video_FaceParser(nn.Module):
    def __init__(self, window_size=7, split_amount=4, img_size=512, folder_name='video_faceparser', ckpt_name = 'raft_things.pth', force=False, device = 'cuda'):
        """
        Related Links
        --------
        https://github.com/williamyang1991/VToonify
        
        https://github.com/zllrunning/face-parsing.PyTorch
        
        
        Keywords
        --------
            - if you setting with 'split amount = 2', Whole frame will be split 2 pieces of chunks.
            - whole frame amount = window size * split amount
            
            |-------------------------------------------- whole frame ---------------------------------------------|
            
            |------------------- window size -----------------|   ...   |------------------ window size ------------------|  
            
            |-- side window --| target frame |-- side window --|   ...   |-- side window --| target frame |-- side window --|  
        
        Label Index
        --------
        
        face_parts | bg| skin | Lbrow | Rbrow | Leye | Reye | glasses | Lear | Rear | ear_ring | nose | mouth | upper_lip | lower_lip | neck | neckless | cloth | hair | hat   
        --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | ---  
        label_index | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18  
        
        Methods
        --------
        - forward
            - image_path_list
                - dtype : list[string]
            - save_path
                - dtype : string
            - vis_save_path
                - Fill this parameters when you want to save visualized label
                - dtype : string
        
        """
        super(Video_FaceParser, self).__init__()
        # parameters
        self.device = device
        self.img_size = 512
        self.window_size = window_size
        self.side_window = self.window_size//2
        self.split_amount = split_amount
    
        self.wt = torch.exp(-(torch.arange(self.window_size).float()-self.side_window)**2/(2*((self.side_window+0.5)**2))).reshape(self.window_size,1,1,1).to(self.device)
    
        url_id = get_url_id('~/.invz_pack/', folder_name, ckpt_name)
        root = os.path.join('~/.invz_pack/', folder_name)
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        ckpt = torch.load(ckpt_path, map_location=self.device)
        
        self.FP = FaceParser(device=self.device)
        
        self.raft = torch.nn.DataParallel(RAFT())
        self.raft.load_state_dict(ckpt)
        self.raft = self.raft.module
        for param in self.raft.parameters():
            param.requires_grad = False
        self.raft.eval().to(self.device)
        del ckpt
    
    def forward(self, frame_path_list, save_path, vis_save_path=None, output_size=512):
        frame_path_list = sorted(frame_path_list)
        os.makedirs(save_path, exist_ok=True)
        if vis_save_path is not None: os.makedirs(vis_save_path, exist_ok=True)
        
        for frame_path_chunk in get_list_chunk(frame_path_list, self.split_amount):
            Is_ = get_batch_frames(frame_path_chunk, self.side_window, self.img_size)

            Ps_ = get_batch_labels(Is_, self.FP, self.side_window)
            # temporal weights of the (2*args.window_size+1) frames
            parse = []
            for ii in tqdm(range(len(frame_path_chunk))):
                i = ii + self.side_window
                image2 = Is_[i-self.side_window:i+self.side_window+1].to(self.device)
                image1 = Is_[i].repeat(2*self.side_window+1,1,1,1).to(self.device)
                padder = InputPadder(image1.shape)
                image1, image2 = padder.pad(image1, image2)
                with torch.no_grad():
                    flow_low, flow_up = self.raft((image1+1)*255.0/2, (image2+1)*255.0/2, iters=20, test_mode=True)
                    output, mask = warp(torch.cat((image2, Ps_[i-self.side_window:i+self.side_window+1].to(self.device)), dim=1), flow_up)
                    
                    aligned_Is = output[:,0:3].detach()
                    aligned_Ps = output[:,3:].detach()
                    
                    # the spatial weight
                    ws = torch.exp(-((aligned_Is-image1)**2).mean(dim=1, keepdims=True)/(2*(0.2**2))) * mask[:,0:1]
                    aligned_Ps[self.side_window] = Ps_[i].to(self.device)
                    # the weight between i and i shoud be 1.0
                    ws[self.side_window,:,:,:] = 1.0
                    weights = ws*self.wt # weights = w(j,i)
                    weights = weights / weights.sum(dim=(0), keepdims=True)
                    fused_Ps = (aligned_Ps * weights).sum(dim=0, keepdims=True)
                    # parse = down(fused_Ps).detach().cpu()
                parse = fused_Ps.detach().cpu().argmax(1) #torch.argmax(torch.cat(parse, dim=0), 1)
                parse = arrange_mask(parse).squeeze().numpy()
                _parse = cv2.resize(parse, (output_size,output_size), interpolation=cv2.INTER_NEAREST)
                image_name = frame_path_chunk[ii].split('/')[-1].split('.')[0]
                cv2.imwrite(os.path.join(save_path, image_name+'.png'), _parse)
                if vis_save_path is not None: cv2.imwrite(os.path.join(vis_save_path, image_name+'.png'), (parse[:,:,None]*10+60)*.5 + (image1[0].detach().permute([1,2,0]).cpu().numpy()[:,:,::-1]*127.5+127.5)*.5)

        print('Done!')



import os
import cv2
import torch
import numpy as np

from torch import nn
import torch.nn.functional as F


def remove_prefix_state_dict(state_dict, prefix="module"):
    """
    remove prefix from the key of pretrained state dict for Data-Parallel
    """
    new_state_dict = {}
    first_state_name = list(state_dict.keys())[0]
    if not first_state_name.startswith(prefix):
        for key, value in state_dict.items():
            new_state_dict[key] = state_dict[key].float()
    else:
        for key, value in state_dict.items():
            new_state_dict[key[len(prefix)+1:]] = state_dict[key].float()
    return new_state_dict



def get_unknown_tensor(trimap):
    """
    get 1-channel unknown area tensor from the 3-channel/1-channel trimap tensor
    """
    weight = trimap[:, 1:2, :, :].float()
    return weight



Kernels = [None] + [cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (size, size)) for size in range(1,30)]


def get_unknown_tensor_from_pred(pred, rand_width=30, train_mode=True):
    ### pred: N, 1 ,H, W 
    N, C, H, W = pred.shape
    _pred = F.interpolate(pred, size=(640,640), mode='nearest')
    _pred = _pred.data.cpu().numpy()
    uncertain_area = np.ones_like(pred, dtype=np.uint8)
    uncertain_area[_pred < 1.0/255.0] = 0
    uncertain_area[_pred > 1-1.0/255.0] = 0

    for n in range(N):
        uncertain_area_ = uncertain_area[n,0,:,:] # H, W
        if train_mode:
            width = np.random.randint(1, rand_width)
        else:
            width = rand_width // 2
        uncertain_area_ = cv2.dilate(uncertain_area_, Kernels[width])
        uncertain_area[n,0,:,:] = uncertain_area_

    weight = np.zeros_like(uncertain_area)
    weight[uncertain_area == 1] = 1

    weight = np.array(weight, dtype=np.float)
    weight = torch.from_numpy(weight).to(pred.device)
    

    weight = F.interpolate(weight, size=(H,W), mode='nearest')

    return weight

class InputPadder:
    """ Pads images such that dimensions are divisible by 8 """
    def __init__(self, dims, mode='sintel'):
        self.ht, self.wd = dims[-2:]
        pad_ht = (((self.ht // 8) + 1) * 8 - self.ht) % 8
        pad_wd = (((self.wd // 8) + 1) * 8 - self.wd) % 8
        if mode == 'sintel':
            self._pad = [pad_wd//2, pad_wd - pad_wd//2, pad_ht//2, pad_ht - pad_ht//2]
        else:
            self._pad = [pad_wd//2, pad_wd - pad_wd//2, 0, pad_ht]

    def pad(self, *inputs):
        return [F.pad(x, self._pad, mode='replicate') for x in inputs]

    def unpad(self,x):
        ht, wd = x.shape[-2:]
        c = [self._pad[2], ht-self._pad[3], self._pad[0], wd-self._pad[1]]
        return x[..., c[0]:c[1], c[2]:c[3]]
    
    
def warp(x, flo):
    """
    warp an image/tensor (im2) back to im1, according to the optical flow
    x: [B, C, H, W] (im2)
    flo: [B, 2, H, W] flow
    """
    B, C, H, W = x.size()
    # mesh grid 
    xx = torch.arange(0, W).view(1,-1).repeat(H,1)
    yy = torch.arange(0, H).view(-1,1).repeat(1,W)
    xx = xx.view(1,1,H,W).repeat(B,1,1,1)
    yy = yy.view(1,1,H,W).repeat(B,1,1,1)
    grid = torch.cat((xx,yy),1).float()


    #x = x.cuda()
    grid = grid.to(x.device)
    vgrid = grid + flo # B,2,H,W

    # scale grid to [-1,1] 
    ##2019 code
    vgrid[:,0,:,:] = 2.0*vgrid[:,0,:,:].clone()/max(W-1,1)-1.0 
    vgrid[:,1,:,:] = 2.0*vgrid[:,1,:,:].clone()/max(H-1,1)-1.0

    vgrid = vgrid.permute(0,2,3,1)
    output = nn.functional.grid_sample(x, vgrid,align_corners=True)
    mask = torch.autograd.Variable(torch.ones(x.size())).to(x.device)
    mask = nn.functional.grid_sample(mask, vgrid,align_corners=True)

    ##2019 author
    mask[mask<0.9999] = 0
    mask[mask>0] = 1

     ##2019 code
     # mask = torch.floor(torch.clamp(mask, 0 ,1))

    return output*mask, mask




Kernels = [None] + [cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (size, size)) for size in range(1,30)]


def get_unknown_tensor_from_pred(pred, rand_width=30, train_mode=True):
    ### pred: N, 1 ,H, W 
    N, C, H, W = pred.shape
    _pred = F.interpolate(pred, size=(640,640), mode='nearest')
    _pred = _pred.data.cpu().numpy()
    uncertain_area = np.ones_like(_pred, dtype=np.uint8)
    uncertain_area[_pred < 1.0/255.0] = 0
    uncertain_area[_pred > 1-1.0/255.0] = 0

    for n in range(N):
        uncertain_area_ = uncertain_area[n,0,:,:] # H, W
        if train_mode:
            width = np.random.randint(1, rand_width)
        else:
            width = rand_width // 2
        uncertain_area_ = cv2.dilate(uncertain_area_, Kernels[width])
        uncertain_area[n,0,:,:] = uncertain_area_

    weight = np.zeros_like(uncertain_area)
    weight[uncertain_area == 1] = 1

    weight = np.array(weight, dtype=np.float)
    weight = torch.from_numpy(weight).to(pred.device)

    weight = F.interpolate(weight, size=(H,W), mode='nearest')

    return weight


import torch
import torch.nn as nn
import torch.nn.functional as F

###### vgg preprocessing ######
def vgg_preprocess(tensor):
    # input is RGB tensor which ranges in [0,1]
    # output is BGR tensor which ranges in [0,255]
    tensor_bgr = torch.cat((tensor[:, 2:3, :, :], tensor[:, 1:2, :, :], tensor[:, 0:1, :, :]), dim=1)
    tensor_bgr_ml = tensor_bgr - torch.Tensor([0.40760392, 0.45795686, 0.48501961]).type_as(tensor_bgr).view(1, 3, 1, 1)
    return tensor_bgr_ml * 255

class VGG19_pytorch(nn.Module):
    """
    NOTE: no need to pre-process the input; input tensor should range in [0,1]
    """

    def __init__(self, pool="max"):
        super(VGG19_pytorch, self).__init__()
        self.conv1_1 = nn.Conv2d(3, 64, kernel_size=3, padding=1)
        self.conv1_2 = nn.Conv2d(64, 64, kernel_size=3, padding=1)
        self.conv2_1 = nn.Conv2d(64, 128, kernel_size=3, padding=1)
        self.conv2_2 = nn.Conv2d(128, 128, kernel_size=3, padding=1)
        self.conv3_1 = nn.Conv2d(128, 256, kernel_size=3, padding=1)
        self.conv3_2 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.conv3_3 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.conv3_4 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.conv4_1 = nn.Conv2d(256, 512, kernel_size=3, padding=1)
        self.conv4_2 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv4_3 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv4_4 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_1 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_2 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_3 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_4 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        if pool == "max":
            self.pool1 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool2 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool3 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool4 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool5 = nn.MaxPool2d(kernel_size=2, stride=2)
        elif pool == "avg":
            self.pool1 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool2 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool3 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool4 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool5 = nn.AvgPool2d(kernel_size=2, stride=2)

    def forward(self, x, out_keys, preprocess=True):
        """
        NOTE: input tensor should range in [0,1]
        """
        out = {}
        if preprocess:
            x = vgg_preprocess(x)
        out["r11"] = F.relu(self.conv1_1(x))
        out["r12"] = F.relu(self.conv1_2(out["r11"]))
        out["p1"] = self.pool1(out["r12"])
        out["r21"] = F.relu(self.conv2_1(out["p1"]))
        out["r22"] = F.relu(self.conv2_2(out["r21"]))
        out["p2"] = self.pool2(out["r22"])
        out["r31"] = F.relu(self.conv3_1(out["p2"]))
        out["r32"] = F.relu(self.conv3_2(out["r31"]))
        out["r33"] = F.relu(self.conv3_3(out["r32"]))
        out["r34"] = F.relu(self.conv3_4(out["r33"]))
        out["p3"] = self.pool3(out["r34"])
        out["r41"] = F.relu(self.conv4_1(out["p3"]))
        out["r42"] = F.relu(self.conv4_2(out["r41"]))
        out["r43"] = F.relu(self.conv4_3(out["r42"]))
        out["r44"] = F.relu(self.conv4_4(out["r43"]))
        out["p4"] = self.pool4(out["r44"])
        out["r51"] = F.relu(self.conv5_1(out["p4"]))
        out["r52"] = F.relu(self.conv5_2(out["r51"]))
        out["r53"] = F.relu(self.conv5_3(out["r52"]))
        out["r54"] = F.relu(self.conv5_4(out["r53"]))
        out["p5"] = self.pool5(out["r54"])
        return [out[key] for key in out_keys]




import cv2
import numpy as np
from PIL import Image

import os, sys, glob
from tqdm import tqdm

import torch
import torch.nn.functional as F

from innerverz import Data_Process
DP = Data_Process()

from innerverz import FaceAligner
FA = FaceAligner(512, align_style='ffhq')

from innerverz import DECA
deca = DECA()

from .util import set_random_colors, plot_kpts

from main import Face_Reshaping
g = Face_Reshaping()

source_list = [
    'YEJI.png',
    'WINTER.png',
    'uc_388.png',
    'uc_648.png',
    'HYERI.jpg',
    'Jenny.jpg',
    '003652.png',
]
video_list = [
    'video_1',
    'video_3',
    'video_4',
    'video_5',
]

save_path = './results_original'

colors = set_random_colors()

from innerverz import FaceParser
FP = FaceParser()        

# source
for source_name in tqdm(source_list):
    source_path = os.path.join(f'./assets/images/{source_name}')
    to_image = cv2.imread(source_path)
    _, to_lmk, _ = FA.detect_lmk(to_image)
    to_face, to_tfm, _, _ = FA.align_face_from_106(to_image, to_lmk)
    align_to_lmk = FA.affine_transform(to_lmk, to_tfm)

    to_face_dict = deca.data_preprocess(to_face, align_to_lmk)
    to_code_dict = deca.encode(to_face_dict['image'][None,...])

    for video_name in tqdm(video_list):
        os.makedirs(f'{save_path}/{source_name[:-4]}_{video_name}', exist_ok=True)
        
        count = 0
        frame_paths = os.path.join(f'./assets/videos/{video_name}/*.*')
        for frame_path in sorted(glob.glob(frame_paths)):
            from_image = cv2.imread(frame_path)
            size = from_image.shape[:-1]
            
            _, from_lmk, _ = FA.detect_lmk(from_image)
            from_face, from_tfm, from_tfm_inv, _ = FA.align_face_from_106(from_image, from_lmk)
            align_from_lmk = FA.affine_transform(from_lmk, from_tfm)
            
            from_face_dict = deca.data_preprocess(from_face, align_from_lmk)
            from_code_dict = deca.encode(from_face_dict['image'][None,...])

            new_code_dict = from_code_dict.copy()
            new_code_dict['shape'] = to_code_dict['shape']

            from_vis_dict = deca.decode(from_code_dict, original_image=from_face_dict['original_image'][None, ...], tform=torch.inverse(from_face_dict['tform'][None, ...]).transpose(1,2))
            from_lmk_points = from_vis_dict['landmarks2d_points']

            new_vis_dict = deca.decode(new_code_dict, original_image=from_face_dict['original_image'][None, ...], tform=torch.inverse(from_face_dict['tform'][None, ...]).transpose(1,2))
            new_lmk_points = new_vis_dict['landmarks2d_points']
            
            # data pp
            from_lmks_vis = plot_kpts(np.zeros_like(from_face), from_lmk_points, colors)
            to_lmks_vis = plot_kpts(np.zeros_like(from_face), new_lmk_points, colors)
            ts_from_face = DP.image_pp(from_face, size=256, batch=True, device='cuda')
            ts_from_lmks_vis = DP.image_pp(from_lmks_vis, size=256, batch=True, device='cuda')
            ts_to_lmks_vis = DP.image_pp(to_lmks_vis, size=256, batch=True, device='cuda')
            
            result = g(ts_from_face, ts_from_lmks_vis, ts_to_lmks_vis)       
            _result = DP.vis_pp(F.interpolate(result,(512,512),mode='bilinear'), normalize=False) # 0 - 1 -> 0 ~ 255
            
            from_label = FP.get_label(F.interpolate(ts_from_face, (512,512)), 512)
            from_innerface_mask = torch.where(from_label < 14, 1, 0) - torch.where(from_label==0, 1, 0) - torch.where(from_label==7, 1, 0) - torch.where(from_label==8, 1, 0) - torch.where(from_label==9, 1, 0)
            _from_innerface_mask = from_innerface_mask.squeeze().cpu().numpy()
            _from_innerface_mask = DP.mask_pp(_from_innerface_mask*255, 5, 0, 15)[...,None] / 255
            
            _from_innerface_mask = cv2.warpAffine(_from_innerface_mask, from_tfm_inv, (size[1], size[0]))
            _result = cv2.warpAffine(_result, from_tfm_inv, (size[1], size[0]))
            
            bg_result = from_image * (1-_from_innerface_mask) + _result * _from_innerface_mask
            # grid = np.concatenate((to_face, from_lmks_vis * .5 + to_lmks_vis * .5, from_face * .5 + from_lmks_vis * .5, _result * .5 + to_lmks_vis * .5, from_face, _result, abs(from_face - _result), bg_result), axis=1)
            grid = np.concatenate((from_image, bg_result), axis=1)
            cv2.imwrite(f'{save_path}/{source_name[:-4]}_{video_name}/{str(count).zfill(6)}.png', grid)
            count += 1
            
        os.system(f'ffmpeg -f image2 -i {save_path}/{source_name[:-4]}_{video_name}/%06d.png -vb 20M -pix_fmt yuv420p -src_range 1 -dst_range 1 {save_path}/{source_name[:-4]}_{video_name}.mp4')


import torch
import torch.nn as nn
import torch.nn.parallel


class Blend_Net(nn.Module):
    def __init__(self, ic):
        super(Blend_Net, self).__init__()
        self.conv1_1 = nn.Sequential(nn.Conv2d(ic, 32, 3, 1, 1), nn.ReLU(), nn.Conv2d(32, 64, 3, 1, 1))
        self.conv1_2 = nn.Conv2d(64, 64, 3, 1, 1)
        self.conv1_2norm = nn.BatchNorm2d(64, affine=False)
        self.conv1_2norm_ss = nn.Conv2d(64, 64, 1, 2, bias=False, groups=64)
        self.conv2_1 = nn.Conv2d(64, 128, 3, 1, 1)
        self.conv2_2 = nn.Conv2d(128, 128, 3, 1, 1)
        self.conv2_2norm = nn.BatchNorm2d(128, affine=False)
        self.conv2_2norm_ss = nn.Conv2d(128, 128, 1, 2, bias=False, groups=128)
        self.conv3_1 = nn.Conv2d(128, 256, 3, 1, 1)
        self.conv3_2 = nn.Conv2d(256, 256, 3, 1, 1)
        self.conv3_3 = nn.Conv2d(256, 256, 3, 1, 1)
        self.conv3_3norm = nn.BatchNorm2d(256, affine=False)
        self.conv3_3norm_ss = nn.Conv2d(256, 256, 1, 2, bias=False, groups=256)
        self.conv4_1 = nn.Conv2d(256, 512, 3, 1, 1)
        self.conv4_2 = nn.Conv2d(512, 512, 3, 1, 1)
        self.conv4_3 = nn.Conv2d(512, 512, 3, 1, 1)
        self.conv4_3norm = nn.BatchNorm2d(512, affine=False)
        self.conv5_1 = nn.Conv2d(512, 512, 3, 1, 2, 2)
        self.conv5_2 = nn.Conv2d(512, 512, 3, 1, 2, 2)
        self.conv5_3 = nn.Conv2d(512, 512, 3, 1, 2, 2)
        self.conv5_3norm = nn.BatchNorm2d(512, affine=False)
        self.conv6_1 = nn.Conv2d(512, 512, 3, 1, 2, 2)
        self.conv6_2 = nn.Conv2d(512, 512, 3, 1, 2, 2)
        self.conv6_3 = nn.Conv2d(512, 512, 3, 1, 2, 2)
        self.conv6_3norm = nn.BatchNorm2d(512, affine=False)
        self.conv7_1 = nn.Conv2d(512, 512, 3, 1, 1)
        self.conv7_2 = nn.Conv2d(512, 512, 3, 1, 1)
        self.conv7_3 = nn.Conv2d(512, 512, 3, 1, 1)
        self.conv7_3norm = nn.BatchNorm2d(512, affine=False)
        self.conv8_1 = nn.ConvTranspose2d(512, 256, 4, 2, 1)
        self.conv3_3_short = nn.Conv2d(256, 256, 3, 1, 1)
        self.conv8_2 = nn.Conv2d(256, 256, 3, 1, 1)
        self.conv8_3 = nn.Conv2d(256, 256, 3, 1, 1)
        self.conv8_3norm = nn.BatchNorm2d(256, affine=False)
        self.conv9_1 = nn.ConvTranspose2d(256, 128, 4, 2, 1)
        self.conv2_2_short = nn.Conv2d(128, 128, 3, 1, 1)
        self.conv9_2 = nn.Conv2d(128, 128, 3, 1, 1)
        self.conv9_2norm = nn.BatchNorm2d(128, affine=False)
        self.conv10_1 = nn.ConvTranspose2d(128, 128, 4, 2, 1)
        self.conv1_2_short = nn.Conv2d(64, 128, 3, 1, 1)
        self.conv10_2 = nn.Conv2d(128, 128, 3, 1, 1)
        self.conv10_ab = nn.Conv2d(128, 3, 1, 1)

        # add self.relux_x
        self.relu1_1 = nn.ReLU()
        self.relu1_2 = nn.ReLU()
        self.relu2_1 = nn.ReLU()
        self.relu2_2 = nn.ReLU()
        self.relu3_1 = nn.ReLU()
        self.relu3_2 = nn.ReLU()
        self.relu3_3 = nn.ReLU()
        self.relu4_1 = nn.ReLU()
        self.relu4_2 = nn.ReLU()
        self.relu4_3 = nn.ReLU()
        self.relu5_1 = nn.ReLU()
        self.relu5_2 = nn.ReLU()
        self.relu5_3 = nn.ReLU()
        self.relu6_1 = nn.ReLU()
        self.relu6_2 = nn.ReLU()
        self.relu6_3 = nn.ReLU()
        self.relu7_1 = nn.ReLU()
        self.relu7_2 = nn.ReLU()
        self.relu7_3 = nn.ReLU()
        self.relu8_1_comb = nn.ReLU()
        self.relu8_2 = nn.ReLU()
        self.relu8_3 = nn.ReLU()
        self.relu9_1_comb = nn.ReLU()
        self.relu9_2 = nn.ReLU()
        self.relu10_1_comb = nn.ReLU()
        self.relu10_2 = nn.LeakyReLU(0.2, True)

        # print("replace all deconv with [nearest + conv]")
        self.conv8_1 = nn.Sequential(nn.Upsample(scale_factor=2, mode="nearest"), nn.Conv2d(512, 256, 3, 1, 1))
        self.conv9_1 = nn.Sequential(nn.Upsample(scale_factor=2, mode="nearest"), nn.Conv2d(256, 128, 3, 1, 1))
        self.conv10_1 = nn.Sequential(nn.Upsample(scale_factor=2, mode="nearest"), nn.Conv2d(128, 128, 3, 1, 1))

        # print("replace all batchnorm with instancenorm")
        self.conv1_2norm = nn.InstanceNorm2d(64)
        self.conv2_2norm = nn.InstanceNorm2d(128)
        self.conv3_3norm = nn.InstanceNorm2d(256)
        self.conv4_3norm = nn.InstanceNorm2d(512)
        self.conv5_3norm = nn.InstanceNorm2d(512)
        self.conv6_3norm = nn.InstanceNorm2d(512)
        self.conv7_3norm = nn.InstanceNorm2d(512)
        self.conv8_3norm = nn.InstanceNorm2d(256)
        self.conv9_2norm = nn.InstanceNorm2d(128)

    def forward(self, x):
        """ x: gray image (1 channel), ab(2 channel), ab_err, ba_err"""
        conv1_1 = self.relu1_1(self.conv1_1(x))
        conv1_2 = self.relu1_2(self.conv1_2(conv1_1))
        conv1_2norm = self.conv1_2norm(conv1_2)
        conv1_2norm_ss = self.conv1_2norm_ss(conv1_2norm)
        conv2_1 = self.relu2_1(self.conv2_1(conv1_2norm_ss))
        conv2_2 = self.relu2_2(self.conv2_2(conv2_1))
        conv2_2norm = self.conv2_2norm(conv2_2)
        conv2_2norm_ss = self.conv2_2norm_ss(conv2_2norm)
        conv3_1 = self.relu3_1(self.conv3_1(conv2_2norm_ss))
        conv3_2 = self.relu3_2(self.conv3_2(conv3_1))
        conv3_3 = self.relu3_3(self.conv3_3(conv3_2))
        conv3_3norm = self.conv3_3norm(conv3_3)
        conv3_3norm_ss = self.conv3_3norm_ss(conv3_3norm)
        conv4_1 = self.relu4_1(self.conv4_1(conv3_3norm_ss))
        conv4_2 = self.relu4_2(self.conv4_2(conv4_1))
        conv4_3 = self.relu4_3(self.conv4_3(conv4_2))
        conv4_3norm = self.conv4_3norm(conv4_3)
        conv5_1 = self.relu5_1(self.conv5_1(conv4_3norm))
        conv5_2 = self.relu5_2(self.conv5_2(conv5_1))
        conv5_3 = self.relu5_3(self.conv5_3(conv5_2))
        conv5_3norm = self.conv5_3norm(conv5_3)
        conv6_1 = self.relu6_1(self.conv6_1(conv5_3norm))
        conv6_2 = self.relu6_2(self.conv6_2(conv6_1))
        conv6_3 = self.relu6_3(self.conv6_3(conv6_2))
        conv6_3norm = self.conv6_3norm(conv6_3)
        conv7_1 = self.relu7_1(self.conv7_1(conv6_3norm))
        conv7_2 = self.relu7_2(self.conv7_2(conv7_1))
        conv7_3 = self.relu7_3(self.conv7_3(conv7_2))
        conv7_3norm = self.conv7_3norm(conv7_3)
        conv8_1 = self.conv8_1(conv7_3norm)
        conv3_3_short = self.conv3_3_short(conv3_3norm)
        conv8_1_comb = self.relu8_1_comb(conv8_1 + conv3_3_short)
        conv8_2 = self.relu8_2(self.conv8_2(conv8_1_comb))
        conv8_3 = self.relu8_3(self.conv8_3(conv8_2))
        conv8_3norm = self.conv8_3norm(conv8_3)
        conv9_1 = self.conv9_1(conv8_3norm)
        conv2_2_short = self.conv2_2_short(conv2_2norm)
        conv9_1_comb = self.relu9_1_comb(conv9_1 + conv2_2_short)
        conv9_2 = self.relu9_2(self.conv9_2(conv9_1_comb))
        conv9_2norm = self.conv9_2norm(conv9_2)
        conv10_1 = self.conv10_1(conv9_2norm)
        conv1_2_short = self.conv1_2_short(conv1_2norm)
        conv10_1_comb = self.relu10_1_comb(conv10_1 + conv1_2_short)
        conv10_2 = self.relu10_2(self.conv10_2(conv10_1_comb))
        conv10_ab = self.conv10_ab(conv10_2)

        return torch.tanh(conv10_ab)
